/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/
#include <firestorm.h>
#include <dotscara/fdctl.h>

#define __USE_XOPEN
#include <unistd.h>
#include <errno.h>
#include <sys/poll.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <features.h>		/* for the glibc version number */
#if __GLIBC__ >= 2 && __GLIBC_MINOR >= 1
#include <net/ethernet.h>	/* the L2 protocols */
#else
#include <asm/types.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>	/* The L2 protocols */
#endif

#include <linux/if.h>
#include <netinet/in.h>
#include <asm/system.h>

#include <plugin.h>
#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <capdev.h>
#include <nbio.h>

static struct proto *sllproto = NULL;

static struct arg linux_args[]={
	{"if", ARGTYPE_STRING, .help = "Interface name"},
	{"blocks", ARGTYPE_UINT, .def = "128",
		.help = "Buffer size in blocks"},
	{"mtu", ARGTYPE_BYTES, .help = "Force MTU"},
	{NULL,}
};

/* This is our own private data */
struct lnx_priv {
	struct packet pkt;
	char *map;
	int num;
	struct tpacket_req req;
	struct iovec *ring;
	size_t maplen;
	int mtu;
	int ifindex;
	struct nbio io;
};

/* Bind us to an interface by index */
static int linux_bindif(int fd, int ifindex)
{
	struct sockaddr_ll addr;

	/* bind the packet socket */
	memset((char *)&addr, 0, sizeof(addr));
	addr.sll_family = AF_PACKET;
	addr.sll_protocol = __constant_htons(0x03);
	addr.sll_ifindex = ifindex;
	addr.sll_hatype = 0;
	addr.sll_pkttype = 0;
	addr.sll_halen = 0;
	if ( bind(fd, (struct sockaddr *)&addr, sizeof(addr)) ) {
		return 0;
	}

	return 1;
}

/* This builds a ring buffer request structure making sure
 * that we have buffers big enough so that a frame which
 * is the size of the MTU doesn't get truncated. We also
 * need to structure things with minimum memory wastage */
static void linux_mkreq(struct lnx_priv *p, unsigned int nr_blocks)
{
	unsigned int pg,bs;
	unsigned int siz;
	unsigned int mult = 1;

	bs = pg = getpagesize();
	siz = p->mtu + TPACKET_HDRLEN;

	if ( bs <= siz ) {
		while( bs < siz ) {
			bs+=pg;
			mult++;
		}

		p->req.tp_block_size = bs;
		p->req.tp_frame_size = bs / mult;
		p->req.tp_block_nr = nr_blocks;
		p->req.tp_frame_nr = mult * nr_blocks;
	}else{
		while( (siz * (mult + 1)) <= pg )
			mult++;

		p->req.tp_block_size = pg;
		p->req.tp_frame_size = pg/mult;
		p->req.tp_block_nr = nr_blocks;
		p->req.tp_frame_nr = mult * nr_blocks;
	}
}

/* Bind the packet socket to a specific interface,
 * we need to lookup the ifindex from the name
 * and we also need to determine the MTU */
static int linux_iface(struct lnx_priv *p, char *ifn)
{
	struct ifreq ifr;
	int ret;

	ret = snprintf(ifr.ifr_name, sizeof(ifr.ifr_name), "%s", ifn);
	if ( ret + 1 > (int)sizeof(ifr.ifr_name) ) {
		mesg(M_ERR, "linux: %s: ifname too big", ifn);
		return 0;
	}

	/* Lookup the MTU */
	if (ioctl(p->io.fd, SIOCGIFMTU, &ifr) < 0) {
		goto out_err;
	}else{
		p->mtu = ifr.ifr_ifru.ifru_mtu;
	}

	/* Lookup the ifindex */
	if (ioctl(p->io.fd, SIOCGIFINDEX, &ifr) < 0) {
		goto out_err;
	}else{
		p->ifindex = ifr.ifr_ifru.ifru_ivalue;
	}

	/* Actually join up to the interface */
	if ( linux_bindif(p->io.fd, p->ifindex) ) {
		return 1;
	}else{
		mesg(M_ERR,"linux: %s: bind(): %s", ifn, sys_err());
		return 0;
	}

out_err:
	mesg(M_ERR,"linux: %s: ioctl(): %s", ifn, sys_err());
	return 0;
}

/* Initialise a capdev process, we open the file and then
 * return our opaque private data structure to firestorm */
static int lnx_init(struct source *s, const char *args)
{
	struct lnx_priv *p;
	unsigned int i;
	char *if_name;
	int ret;

	p = calloc(1, sizeof(struct lnx_priv));
	if ( p == NULL )
		goto err;

	if ( !sllproto && !(sllproto = decode_proto("linux")) )
		goto err_free;

	/* Open the packet socket */
	if ( (p->io.fd = socket(PF_PACKET, SOCK_DGRAM, 0))<0 ) {
		mesg(M_ERR,"linux: socket(): %s", sys_err());
		goto err_free;
	}

	/* Parse args */
	switch ( args_parse(linux_args, args, NULL) ) {
	case -1:
		mesg(M_ERR,"linux: parse error: %s", args);
	case 0: /* fall through */
		goto err_free;
	default:
		break;
	}

	if_name = arg_get_string(&linux_args[0].val.v_str);
	if ( if_name == NULL )
		if_name = "any";

	/* Join interfaces */
	if ( !strcmp(if_name, "any") ) {
		p->ifindex = 0; /* zero == all */
		if ( !linux_bindif(p->io.fd, p->ifindex) ) {
			mesg(M_ERR,"linux: %s: bind(): %s",
				if_name, sys_err());
			goto err_close;
		}

		/* TODO: enumerate all interfaces for
		 * highest possible MTU */
		if ( !linux_args[2].val.v_bytes )
			p->mtu = 2048 - TPACKET_HDRLEN;
		else
			p->mtu = linux_args[2].val.v_bytes;
	}else{
		if ( linux_args[2].val.v_bytes ) {
			mesg(M_WARN,"linux: You are already specifying an "
				"interface no need to force MTU!");
		}

		if ( !linux_iface(p, if_name) )
			goto err_close;
	}

	/* Setup socket for RX ring buffer */
	linux_mkreq(p, linux_args[1].val.v_uint);
	ret = setsockopt(p->io.fd,SOL_PACKET,PACKET_RX_RING,
			(char *)&p->req, sizeof(p->req));
	if ( ret ) {
		if ( errno == ENOPROTOOPT )
			mesg(M_ERR, "linux: you need CONFIG_PACKET_MMAP "
				"in your kernel");
		else
			mesg(M_ERR,"linux: setsockopt(): %s", sys_err());
		goto err_close;
	}

	/* mmap() the sucker */
	p->maplen = p->req.tp_block_size * p->req.tp_block_nr;
	p->map = mmap(NULL, p->maplen,
				PROT_READ|PROT_WRITE|PROT_EXEC,
				MAP_SHARED, p->io.fd, 0);
	if ( p->map == MAP_FAILED ) {
		mesg(M_ERR,"linux: mmap(): %s", sys_err());
		goto err_close;
	}

	/* Allocate ring buffer iovecs */
	if ( !(p->ring = malloc(p->maplen)) ) {
		mesg(M_ERR,"linux: malloc(): %s", sys_err());
		goto err_unmap;
	}

	/* Setup our ring buffer */
	for(i = 0; i<p->req.tp_frame_nr; i++) {
		p->ring[i].iov_base=(void *)(p->map+(i*p->req.tp_frame_size));
		p->ring[i].iov_len = p->req.tp_frame_size;
	}

	/* Print information */
	mesg(M_INFO,"linux: %s(%i): mtu=%i blocks=%u buffer=%uKB (%u frames)",
		if_name, p->ifindex, p->mtu, linux_args[1].val.v_uint,
		(p->req.tp_block_size*p->req.tp_block_nr)/1024,
		p->req.tp_frame_nr);

	capdev_init_packet(&p->pkt, s);

	s->priv = p;

	p->io.mask = NBIO_READ;
	p->io.priv = s;
	capdev_fd_add(s, &p->io);

	return 1;

err_unmap:
	munmap(p->map, p->maplen);
err_close:
	fd_close(p->io.fd);
err_free:
	free(p);
err:
	return 0;
}

static void lnx_free(struct source *s)
{
	struct lnx_priv *p = s->priv;
	struct tpacket_stats st;
	int len = sizeof(st);

	/* Print statistics */
	if ( !getsockopt(p->io.fd, SOL_PACKET,
		PACKET_STATISTICS, (char *)&st, &len)) {
		mesg(M_INFO,"linux: received %u packets, dropped %u",
			st.tp_packets, st.tp_drops);
	}

	munmap(p->map, p->maplen);
	fd_close(p->io.fd);
	free(p);
}

static struct packet *dequeue(struct source *s)
{
	struct lnx_priv *p = s->priv;
	struct tpacket_hdr *h;
	struct sockaddr_ll *sll;
	struct timeval tv;

	h = p->ring[p->num].iov_base;
	if ( h->tp_status == TP_STATUS_KERNEL ) {
		capdev_fd_inactive(s, &p->io);
		return NULL;
	}

	sll = (void *)h + TPACKET_ALIGN(sizeof(*h));

	/* build packet */
	tv.tv_sec = h->tp_sec;
	tv.tv_usec = h->tp_usec;
	p->pkt.time = time_from_timeval(&tv);
	p->pkt.len = h->tp_mac+h->tp_len;
	p->pkt.caplen = h->tp_mac+h->tp_snaplen;
	p->pkt.base = h;
	p->pkt.end = (void *)h + p->pkt.caplen;

	/* Decode */
	p->pkt.layer[0].proto = sllproto;
	p->pkt.layer[0].h.raw = sll;
	p->pkt.layer[0].flags = 0;
	p->pkt.layer[0].session = NULL;
	p->pkt.llen = 0;
	p->pkt.alert = NULL;

	p->pkt.flags = FP_COPY|FP_REAL|FP_PKTTYPE|
			FP_WRITABLE|capdev_endian_flag(0);

	switch ( sll->sll_pkttype ) {
	case 0:
		p->pkt.flags |= FP_HOST;
		break;
	case 1:
		p->pkt.flags |= FP_BROADCAST;
		break;
	case 2:
		p->pkt.flags |= FP_MULTICAST;
		break;
	case 4:
		p->pkt.flags |= FP_OUTGOING;
		break;
	case 3:
	default:
		p->pkt.flags |= FP_PROMISC;
		break;
	}

	/* move along the ring buffer */
	h->tp_status = 0;
	mb();
	p->num = (p->num + 1) % p->req.tp_frame_nr;

	return &p->pkt;
}

struct capdev firestorm_capdevs[]={
	{
		.name = "linux",
		.desc = "Fast live capture for Linux",
		.init = lnx_init,
		.free = lnx_free,
		.dequeue = dequeue,
		.flags = CAPDEV_NBIO|CAPDEV_REALTIME,
		.args = linux_args,
	},
	{ .name = NULL }
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "capdev.linux",
	.desc		= "Linux mmap() packet socket()",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 1,
	.ver_minor	= 0,
};
