/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/
#include <firestorm.h>
#include <plugin.h>
#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <capdev.h>
#include <args.h>
#include <nbio.h>
#include <pcap.h>

#define READ_TIMEOUT 500

struct fpcap_priv {
	pcap_t		*pcap_desc;
	struct packet	pkt;
	struct proto	*proto;
	char		*ifname;
	struct nbio	io;
};

static struct arg pl_args[]={
	{"if", ARGTYPE_STRING, .help = "Interface name"},
	{"mtu", ARGTYPE_BYTES, .def = "16384", .help = "Force MTU"},
	{"promisc", ARGTYPE_BOOL,
		.help = "Set promiscuous mode during capture"},
	{NULL}
};

static struct arg pf_args[]={
	{"file", ARGTYPE_STRING, .help = "Input filename"},
	{NULL}
};

static int finish_init(struct fpcap_priv *p, struct source *s, int live)
{
	unsigned int lnk;
	
	lnk = pcap_datalink(p->pcap_desc);

	p->proto = decode_subproto("__pcap_dlt", lnk);
	if ( p->proto == NULL ) {
		mesg(M_ERR,"pcap: %s: Can't support protocol 0x%x",
			p->ifname, lnk);
		return 0;
	}

	capdev_init_packet(&p->pkt, s);
	return 1;
}

static int file_init(struct source *s, const char *args)
{
	char ebuf[PCAP_ERRBUF_SIZE];
	struct fpcap_priv *p;
	char *ifname;

	switch( args_parse(pf_args, args, NULL) ) {
	case -1:
		mesg(M_ERR, "pcap: parse error: %s", args);
	case 0: /* fall through */
		return 0;
	default:
		break;
	}

	ifname = arg_get_string(&pf_args[0].val.v_str);
	if ( ifname == NULL ) {
		mesg(M_ERR, "pcap: no file specified: %s", args);
		return 0;
	}

	p = calloc(1, sizeof(*p));
	if ( p == NULL )
		return 0;

	p->ifname = strdup(ifname);
	if ( p->ifname == NULL ) {
		mesg(M_ERR, "pcap: %s: strdup(): %s", ifname, sys_err());
		goto err;
	}

	p->pcap_desc = pcap_open_offline(p->ifname, ebuf);
	if ( p->pcap_desc == NULL ) {
		mesg(M_ERR,"pcap: %s", ebuf);
		goto err;
	}

	if ( !finish_init(p, s, 0) )
		goto err;

	s->priv = p;
	return 1;
err:
	if ( p->ifname )
		free(p->ifname);
	free(p);
	return 0;
}

static int live_init(struct source *s, const char *args)
{
	char ebuf[PCAP_ERRBUF_SIZE];
	struct fpcap_priv *p;
	char *ifname;

	switch( args_parse(pl_args, args, NULL) ) {
	case -1:
		mesg(M_ERR,"pcap: parse error: %s", args);
	case 0: /* fall through */
		return 0;
	default:
		break;
	}

	ifname = arg_get_string(&pl_args[0].val.v_str);
	if ( ifname == NULL ) {
		mesg(M_ERR, "pcap: no interface specified: %s", args);
		return 0;
	}

	p = calloc(1, sizeof(*p));
	if ( p == NULL )
		return 0;

	p->ifname = strdup(ifname);
	if ( p->ifname == NULL ) {
		mesg(M_ERR, "pcap: %s: strdup(): %s", ifname, sys_err());
		goto err;
	}

	ebuf[0] = '\0';
	p->pcap_desc = pcap_open_live(p->ifname,
					pl_args[1].val.v_bytes, /* mtu */
					pl_args[2].val.v_bool, /* promisc */
					READ_TIMEOUT, ebuf);
	if ( p->pcap_desc == NULL ) {
		mesg(M_ERR,"pcap: %s", ebuf);
		goto err;
	}

	if ( ebuf[0] != '\0' )
		mesg(M_WARN, "pcap: %s", ebuf);

	if ( !finish_init(p, s, 0) )
		goto err;

	if ( pcap_setnonblock(p->pcap_desc, 1, ebuf) ) {
		mesg(M_ERR, "pcap: %s", ebuf);
		goto err;
	}

	s->priv = p;

	p->io.fd = pcap_fileno(p->pcap_desc);
	p->io.mask = NBIO_READ;
	p->io.priv = s;
	capdev_fd_add(s, &p->io);

	return 1;
err:
	if ( p->ifname )
		free(p->ifname);
	free(p);
	return 0;
}
static void pcap_free(struct source *s)
{
	struct fpcap_priv *p = s->priv;
	struct pcap_stat stat;

	if ( p == NULL )
		return;

        if (!pcap_file(p->pcap_desc) && pcap_stats(p->pcap_desc, &stat) < 0) {
		mesg(M_ERR, "pcap: pcap_stats(): %s",
			pcap_geterr(p->pcap_desc));
        }else{
		mesg(M_INFO,"pcap: received %u packets, dropped %u",
			stat.ps_recv, stat.ps_drop);
	}

	if ( p->pcap_desc )
		pcap_close(p->pcap_desc);

	if ( p->ifname )
		free(p->ifname);

	free(p);
}

static void lpf_callback(u_char *user, struct pcap_pkthdr *header, u_char *data)
{
	struct fpcap_priv *p;

	p = (struct fpcap_priv *)user;
	if ( p == NULL )
		return;

	/* Pointers */
	p->pkt.base = data;
	p->pkt.end = data;
	p->pkt.end += header->caplen;

	/* length */
	p->pkt.caplen = header->caplen;
	p->pkt.len = header->len;

	/* Timestamp */
	p->pkt.time = time_from_timeval(&header->ts);

	/* first layer */
	p->pkt.layer[0].proto = p->proto;
	p->pkt.layer[0].h.raw = p->pkt.base;
	p->pkt.layer[0].flags = 0;
	p->pkt.layer[0].session = NULL;
	p->pkt.llen = 0;
	p->pkt.alert = NULL;

	p->pkt.flags = FP_COPY|FP_REAL|FP_PROMISC|FP_WRITABLE|
			capdev_endian_flag(pcap_is_swapped(p->pcap_desc));
}

static struct packet *live_dequeue(struct source *s)
{
	struct fpcap_priv *p = s->priv;
	int ret;
	
	ret = pcap_dispatch(p->pcap_desc, -1,
		(pcap_handler)lpf_callback, s->priv);

	if ( ret < 0 ) {
		mesg(M_ERR, "pcap: %s", pcap_geterr(p->pcap_desc));
		capdev_fd_inactive(s, &p->io);
		return NULL;
	}

	if ( ret == 0 ) {
		capdev_fd_inactive(s, &p->io);
		return NULL;
	}

	return &p->pkt;
}

static struct packet *file_dequeue(struct source *s)
{
	struct fpcap_priv *p = s->priv;
	int ret;
	
	ret = pcap_dispatch(p->pcap_desc, 1,
		(pcap_handler)lpf_callback, s->priv);

	if ( ret < 0 ) {
		mesg(M_ERR, "pcap: %s", pcap_geterr(p->pcap_desc));
		return NULL;
	}

	if ( ret == 0 )
		return NULL;

	return &p->pkt;
}

struct capdev firestorm_capdevs[]={
	{
		.name = "pcap",
		.desc = "libpcap live capture",
		.init = live_init,
		.free = pcap_free,
		.dequeue = live_dequeue,
		.flags = CAPDEV_NBIO|CAPDEV_REALTIME,
		.args = pl_args,
	},
	{
		.name = "pcapfile",
		.desc = "libpcap offline capture",
		.init = file_init,
		.free = pcap_free,
		.dequeue = file_dequeue,
		.args = pf_args,
	},
	{.name = NULL}
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "capdev.pcap",
	.desc		= "libpcap capture",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 1,
	.ver_minor	= 0,
};
