/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/

#include <firestorm.h>

#include <sys/socket.h>
#include <netinet/in.h>

#include <args.h>
#include <plugin.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>

#define DLT_EN10MB 1

void ether_decode(struct packet *);
void null_decode(struct packet *);
void llc_decode(struct packet *);
void snap_decode(struct packet *);
int ether_dprint(struct packet*, int, char *, size_t);
int llc_dprint(struct packet *, int, char *, size_t);
int snap_dprint(struct packet *, int, char *, size_t);

/*
 * These are to be registered using the "decode.add" function.
 * These structures cannot be declared const, they must also be
 * initialised with the macros provided.
 */
struct proto ether_p=init_proto("ethernet", ether_decode, ether_dprint);

/*
 * These structures tell firestorm what relation our
 * protocols have to existing protocols. They must
 * be initialised with the provided macros. Arrays
 * must be terminated with a null element.
 */
struct proto_req ether_r[]={
	proto_request("__pcap_dlt", DLT_EN10MB),
	null_request()
};
struct proto_req snap_r[]={
	proto_request("llc", 0xaa),
	null_request()
};

/* 802.3 MAC hacks */
struct proto_req mac_r[]={null_request()};
struct proto mac_p=init_proto("802.3", null_decode, ether_dprint);
struct proto nw_p=init_proto("802.3-novell", null_decode, ether_dprint);
struct proto llc_p=init_proto("llc", llc_decode, llc_dprint);
struct proto snap_p=init_proto("snap", snap_decode, snap_dprint);

/* Print out a description of an ethernet header */
int ether_dprint(struct packet *p, int l, char *buf, size_t buflen)
{
	char *b = buf;
	char *x = p->layer[l].h.raw + 6;
	u_int16_t proto;
	int c;

	/* SRC */
	for(c=0; c<6; c++, x++)
		b+=sprintf(b, "%02x%c", *x&0xFF, c==5 ? ' ' : ':');

	b+=sprintf(b, "> ");

	/* DST */
	x = p->layer[l].h.raw;
	for(c=0; c<6; c++, x++)
		b+=sprintf(b, "%02x%c", *x&0xFF, c==5 ? ' ' : ':');

	proto=ntohs(p->layer[l].h.eth->proto);
	if ( proto <= 1500 ) {
		b+=sprintf(b, "length=%u", proto);
	}else{
		b+=sprintf(b, "proto=0x%.4x", proto);
	}

	return buflen;
}

int llc_dprint(struct packet *p, int l, char *buf, size_t buflen)
{
	return snprintf(buf, buflen, "dsap=0x%x lsap=0x%x",
		p->layer[l].h.llc->dsap & 0xff,
		p->layer[l].h.llc->lsap & 0xff);
}

/* LLC Decoding, usually a SNAP header next */
void llc_decode(struct packet *p)
{
	struct proto *child;
	struct layer *l=&p->layer[p->llen];

	if ( (p->layer[p->llen+1].h.raw=
		l->h.raw+sizeof(struct pkt_llchdr))
		> p->end ) goto finish;

	p->llen++;

	if ( p->llen >= PKT_LAYERS ) {
		dispatch(p);
		return;
	}

	/* FIXME: Is checking the DSAP correct behaivour? */
	child = decode_next(l->proto, l->h.llc->dsap);
	if ( child ) {
		p->layer[p->llen].flags = 0;
		p->layer[p->llen].session = NULL;
		p->layer[p->llen].proto = child;
		child->decode(p);
		return;
	}

finish:
	if ( p->layer[p->llen].h.raw<p->end )
		p->layer[p->llen++].proto=NULL;
	dispatch(p);
}

int snap_dprint(struct packet *p, int l, char *buf, size_t buflen)
{
	struct layer *a = &p->layer[l];
	return snprintf(buf, buflen, "org=%.2x%.2x%.2x",
		a->h.snap->org[0],
		a->h.snap->org[1],
		a->h.snap->org[2]);
}
void snap_decode(struct packet *p)
{
	struct proto *child;
	struct layer *l=&p->layer[p->llen];

	/* Fill in the next layers pointer */
	if ( (p->layer[p->llen+1].h.raw=
		l->h.raw+sizeof(struct pkt_snaphdr))
		> p->end ) goto finish;

	p->llen++;

	if ( p->llen >= PKT_LAYERS ) {
		dispatch(p);
		return;
	}

	/* SNAP has the same children as ethernet II */
	child = decode_next(&ether_p, l->h.snap->proto);
	if ( child ) {
		p->layer[p->llen].flags = 0;
		p->layer[p->llen].session = NULL;
		p->layer[p->llen].proto = child;
		child->decode(p);
		return;
	}

finish:
	if ( p->layer[p->llen].h.raw<p->end )
		p->layer[p->llen++].proto=NULL;
	dispatch(p);
}

void null_decode(struct packet *p){
	mesg(M_DEBUG,"ether: null decode called!");
	return;
}

/* This function actually does the decoding */
void ether_decode(struct packet *p)
{
	struct proto *child;
	struct layer *l=&p->layer[p->llen];
	int i=p->llen;

	/* Fill in the next layers pointer */
	if ( (p->layer[p->llen+1].h.raw=
		l->h.raw+sizeof(struct pkt_ethhdr))
		> p->end ) goto finish;

	p->llen++;

	/* Check if we are 802.3 and change accordingly */
	if ( ntohs(l->h.eth->proto)<=1500 ) {
		/* Ugly hack for Novell Netware IPX frames */
		if ( (p->layer[i].h.raw+2 < p->end) &&
			((unsigned char *)p->layer[i+1].h.raw)[0]==0xff &&
			((unsigned char *)p->layer[i+1].h.raw)[1]==0xff ) {
			l->proto=&nw_p;
		}else{
			/* We are 802.3 and so the next header
			 * will always be LLC */
			l->proto=&mac_p;
			if ( p->llen >= PKT_LAYERS ) {
				dispatch(p);
				return;
			}
			p->layer[p->llen].flags=0;
			p->layer[p->llen].session=NULL;
			p->layer[p->llen].proto=&llc_p;
			llc_p.decode(p);
			return;
		}
	}

	if ( p->llen >= PKT_LAYERS ) {
		dispatch(p);
		return;
	}

	child = decode_next(l->proto, l->h.eth->proto);
	if ( child ) {
		p->layer[p->llen].flags=0;
		p->layer[p->llen].session=NULL;
		p->layer[p->llen].proto=child;
		child->decode(p);
		return;
	}

	/* Just data */
finish:
	if ( p->layer[p->llen].h.raw<p->end )
		p->layer[p->llen++].proto=NULL;
	dispatch(p);
}

struct proto_table firestorm_protocols[]={
	{.proto = &ether_p, .preq = ether_r},
	{.proto = &mac_p, .preq = mac_r},
	{.proto = &nw_p, .preq = mac_r},
	{.proto = &llc_p, .preq = mac_r},
	{.proto = &snap_p, .preq = snap_r},
	{.proto = NULL}
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "decode.ether",
	.desc		= "Ethernet II, 802.3, LLC and SNAP",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 2,
	.ver_minor	= 1,
};
