/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * Released under the terms of the GNU GPL version 2
*/
#include <firestorm.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "http_normalize.h"

#define HTTP_NEXGEN_MAGIC	0xffff
#define HTTP_NEXGEN_VERSION	0
struct http_s_req {
	u_int16_t magic;
	u_int16_t version;
	u_int32_t hdr_len;

	u_int8_t type;
	u_int8_t proto_vers;
	u_int16_t port;

	u_int32_t method_ofs;
	u_int32_t method_len;
	u_int32_t host_ofs;
	u_int32_t host_len;
	u_int32_t uri_ofs;
	u_int32_t uri_len;

	u_int32_t path_len;
	u_int32_t query_len;

	/* Normalized path */
	/* Normalized query string */
};

struct http_serial_resp {
	u_int8_t type;
	u_int8_t vers;
	u_int16_t unused;
	int32_t code;
};

/* For backwards compat */
struct http_s_v1 {
	u_int32_t uri;
	u_int32_t uri_len;
	u_int32_t method;
	u_int32_t method_len;
};

struct http_s_req_v2 {
	u_int8_t type;
	u_int8_t vers;
	u_int16_t unused;
	u_int32_t uri;
	u_int32_t uri_len;
	u_int32_t method;
	u_int32_t method_len;
};

int __private http_serialize(struct packet *p, int i, char *buf, size_t len)
{
	struct layer *l=&p->layer[i];
	struct http_request *s=(struct http_request *)l->session;
	struct http_response *sr=(struct http_response *)l->session;
	struct http_s_req *f=(struct http_s_req *)buf;
	struct http_serial_resp *r=(struct http_serial_resp *)buf;
	int tot_len;
	void *ptr;

	if ( s->type == HTTP_REQUEST ) {
		if ( len < sizeof(*f) )
			return -1;

		tot_len = sizeof(*f) + s->uri_path.len + s->uri_query.len;
		if ( tot_len < 0 || (size_t)tot_len > len )
			return -1;

		memset(buf, 0, tot_len);

		f->magic = htons(HTTP_NEXGEN_MAGIC);
		f->version = htons(HTTP_NEXGEN_VERSION);
		f->hdr_len = htonl(sizeof(*f));
		f->type = HTTP_REQUEST;
		f->proto_vers = s->proto_vers;
		f->port = htons(s->port);

		/* Pointers to parts of the packet */
		if ( s->method.ptr ) {
			f->method_ofs = htonl(s->method.ptr - l->h.raw);
			f->method_len = htonl(s->method.len);
		}
		if ( s->host.ptr ) {
			f->host_ofs = htonl(s->host.ptr - l->h.raw);
			f->host_len = htonl(s->host.len);
		}
		if ( s->uri.ptr ) {
			f->uri_ofs = htonl(s->uri.ptr - l->h.raw);
			f->uri_len = htonl(s->uri.len);
		}

		/* Out-of-band normalized stuff */
		f->path_len = htonl(s->uri_path.len);
		f->query_len = htonl(s->uri_query.len);

		ptr = buf + sizeof(*f);
		memcpy(ptr, s->uri_path.ptr, s->uri_path.len);
		ptr += s->uri_path.len;
		memcpy(ptr, s->uri_query.ptr, s->uri_query.len);
		ptr += s->uri_query.len;

		return tot_len;
	}else{
		if ( len < sizeof(*r) )
			return -1;
		r->type = sr->type;
		r->vers = sr->proto_vers;
		r->unused = 0;
		r->code = htonl(sr->code);
		return sizeof(*f);
	}
}

/* For backwards compatibility */
static void *http_deserialize_v1(struct packet *p, int i, const char *buf)
{
	struct http_s_v1 *f = (struct http_s_v1 *)buf;
	struct http_request *s;
	struct layer *l=&p->layer[i];

	s = calloc(1, sizeof(*s));
	if ( s == NULL )
		return NULL;

	s->type = HTTP_REQUEST;
	s->proto_vers = HTTP_VER_UNKNOWN;
	s->uri.ptr = l->h.raw + ntohl(f->uri);
	s->uri.len = ntohl(f->uri_len);
	s->method.ptr = l->h.raw + ntohl(f->method);
	s->method.len = ntohl(f->method_len);
	s->port = HTTP_DEFAULT_PORT;

	return s;
}

static void *http_deserialize_v2(struct packet *p, int i, const char *buf)
{
	struct http_s_req_v2 *f = (struct http_s_req_v2 *)buf;
	struct http_request *s;
	struct layer *l=&p->layer[i];
	u_int32_t plen;

	plen=(char *)p->end - (char *)l->h.raw;

	if ( !(s=calloc(1, sizeof(*s))) )
		return NULL;

	if ( plen < ntohl(f->uri) + ntohl(f->uri_len) )
		goto fail;
	if ( plen < ntohl(f->method) + ntohl(f->method_len) )
		goto fail;

	s->type = f->type;
	s->proto_vers = f->vers;
	s->uri.ptr = l->h.raw + ntohl(f->uri);
	s->uri.len = ntohl(f->uri_len);
	s->method.ptr = l->h.raw + ntohl(f->method);
	s->method.len = ntohl(f->method_len);
	s->port = HTTP_DEFAULT_PORT;

	return s;
fail:
	free(s);
	return NULL;
}

static void *http_nexgen_deserialize(struct packet *p, int i,
				const char *buf, size_t len)
{
	struct http_s_req *cur = (struct http_s_req *)buf;
	struct layer *l = &p->layer[i];
	struct http_request *s;
	const char *str;
	size_t strsz;

	if ( !(s=calloc(1, sizeof(*s))) )
		return NULL;

	if ( cur->type != HTTP_REQUEST )
		goto fail;

	if ( ntohs(cur->version) != HTTP_NEXGEN_VERSION )
		goto fail;

	if ( ntohl(cur->hdr_len) > len )
		goto fail;

	str = buf + ntohl(cur->hdr_len);

	s->type = cur->type;
	s->proto_vers = cur->proto_vers;
	s->port = ntohs(cur->port);

	s->method.ptr = l->h.raw + ntohl(cur->method_ofs);
	s->method.len = ntohl(cur->method_len);
	s->host.ptr = l->h.raw + ntohl(cur->host_ofs);
	s->host.len = ntohl(cur->method_len);
	s->uri.ptr = p->layer[i].h.raw + ntohl(cur->uri_ofs);
	s->uri.len = ntohl(cur->uri_len);

	if ( (s->method.ptr + s->method.len > p->end) )
		goto fail;
	if ( (s->host.ptr + s->host.len > p->end) )
		goto fail;

	strsz = len - ntohl(cur->hdr_len);
	if ( strsz == 0 )
		goto done;

	s->free = malloc(strsz);
	memcpy(s->free, str, strsz);
	s->uri_path.len = ntohl(cur->path_len);
	s->uri_query.len = ntohl(cur->query_len);
	s->uri_path.ptr = s->free;
	s->uri_query.ptr = s->free + s->uri_path.len;


done:
	return s;
fail:
	free(s);
	return NULL;
}

/* Remember: this data is un-trusted so validate it! */
void __private *http_deserialize(struct packet *p, int i,
				const char *buf, size_t len)
{
	struct http_request *s;
	struct http_response *sr;
	struct http_s_req_v2 *f=(struct http_s_req_v2 *)buf;
	struct http_serial_resp *r=(struct http_serial_resp *)buf;
	struct http_s_req *cur = (struct http_s_req *)buf;
	
	if ( len == sizeof(struct http_s_v1) )
		return http_deserialize_v1(p, i, buf);

	if ( len >= sizeof(struct http_s_req) &&
		(ntohs(cur->magic) == HTTP_NEXGEN_MAGIC) ) {
		return http_nexgen_deserialize(p, i, buf, len);
	}

	switch ( f->type ) {
	case HTTP_REQUEST:
		if ( len == sizeof(*f) )
			return http_deserialize_v2(p, i, buf);
		return NULL;

	case HTTP_RESPONSE:
		if ( len != sizeof(*r) )
			return NULL;
		
		if ( !(s=calloc(1, sizeof(*sr))) )
			return NULL;

		sr = (struct http_response *)s;
		sr->type = r->type;
		sr->proto_vers = r->vers;
		sr->code = (int)ntohl(r->code);

		return sr;

	default:
		return NULL;
	}
}
