/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/
#include <firestorm.h>

#include <sys/socket.h>
#include <netinet/in.h>

#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <idx.h>
#include <field.h>
#include <matcher.h>
#include <plugin.h>
#include <strtouint.h>

static struct proto *tcpproto = NULL;
static struct proto *udpproto = NULL;

struct rpc_pkt {
	u_int32_t xid;
	u_int16_t direction;
	u_int32_t rpc_ver;
	u_int32_t app, ver, proc;
};

#define RPC_CHECK_APP	(1<<0)
#define RPC_CHECK_PROC	(1<<1)
#define RPC_CHECK_VER	(1<<2)
struct rpc_node {
	u_int32_t app, ver, proc;
	u_int32_t flags;
	int negate;
	struct alert *alert;
	struct detect *result;
	struct rpc_node *next;
};

struct rpc_detect {
	struct detect d;
	struct rpc_node *nodes;
};

static const u_int16_t rpc_call;
static const u_int32_t rpc_ver=__constant_htonl(0x0002);

static inline int rpc_matches(struct rpc_pkt *rp, struct rpc_node *n)
{
	if ( (n->flags & RPC_CHECK_APP) && (n->app != rp->app) )
		return 0;
	if ( (n->flags & RPC_CHECK_VER) && (n->ver != rp->ver) )
		return 0;
	if ( (n->flags & RPC_CHECK_PROC) && (n->proc != rp->proc) )
		return 0;
	return 1;
}

static void rpc_match(struct detect *d, struct packet *p, int l)
{
	struct rpc_detect *x = (struct rpc_detect *)d;
	struct rpc_pkt *rp;
	struct rpc_node *n;

	/* Find RPC packet header */
	if ( p->layer[l+1].proto == tcpproto ) {
		rp = (struct rpc_pkt *)(p->layer[l+2].h.raw + 4);
	}else if ( p->layer[l+1].proto == udpproto ) {
		rp = (struct rpc_pkt *)p->layer[l+2].h.raw;
	}else{
		goto skip;
	}

	if ( (void *)rp + sizeof(*rp) > p->end )
		goto skip;

	if ( rp->direction != rpc_call )
		goto skip;

	if ( rp->rpc_ver != rpc_ver )
		goto skip;

	for(n=x->nodes; n; n=n->next) {
		int ret = rpc_matches(rp, n);

		if ( n->negate ^ ret ) {
			if ( n->alert )
				alert_tag(p, n->alert, l);
			if ( n->result )
				n->result->match(n->result, p, l);
		}
	}

skip:
	if ( x->d.any )
		x->d.any->match(x->d.any, p, l);
}

static struct detect *rpc_new(struct matcher *m)
{
	struct rpc_detect *x;

	x = calloc(1, sizeof(*x));
	if ( x == NULL )
		return NULL;

	x->d.match = rpc_match;

	return (struct detect *)x;
}

static struct detect **rpc_insert(struct detect *d, struct sig_node *n)
{
	struct rpc_detect *x = (struct rpc_detect *)d;
	struct rpc_node *p = (struct rpc_node *)n->private;
	struct rpc_node *tmp;

	p->alert = n->alert;
	p->negate = n->negate;

	for(tmp=x->nodes; tmp; tmp=tmp->next) {
		if ( p->app != tmp->app )
			continue;
		if ( p->ver != tmp->ver )
			continue;
		if ( p->proc != tmp->proc )
			continue;
		if ( p->flags != tmp->flags )
			continue;
		if ( p->negate != tmp->negate )
			continue;

		tmp->alert = alert_best(tmp->alert, p->alert);
		n->match->free_priv(n->private);
		return &tmp->result;
	}

	p->next = x->nodes;
	x->nodes = p;
	
	return &p->result;
}

static int rpc_validate(struct matcher *match, char *args,
			void **priv, struct modifier *m)
{
	struct rpc_node *p;
	u_int32_t app=0, proc=0, ver=0;
	u_int32_t flags=0;
	char *str[3];
	int i;

	if ( !args )
		return 0;

	/* We only support RPC encapsulations in TCP or UDP */
	if ( !tcpproto && !(tcpproto=decode_proto("tcp")) )
		return 0;

	if ( !udpproto && !(udpproto=decode_proto("udp")) )
		return 0;

	str[0]=args;
	str[1]=NULL;
	str[2]=NULL;
	for(i=0; *args; args++) {
		if ( i==0 ) {
			if ( !isspace(*args) )
				str[i++]=args;
		}else if ( i<3 ) {
			if ( *args==',' ) {
				*args=0;
				str[i++]=args+1;
			}
		}else{
			if ( isspace(*args) ) {
				*args=0;
				break;
			}
		}
	}

	if ( i<3 )
		return 0;

	/* Application / Program */
	if ( str[0][0]!='*' ) {
		if ( strtouint(str[0], &app) )
			return 0;
		flags|=RPC_CHECK_APP;
	}

	/* Procedure */
	if ( str[1][0]!='*' ) {
		if ( strtouint(str[1], &proc) )
			return 0;
		flags|=RPC_CHECK_PROC;
	}

	/* Version */
	if ( str[2][0]!='*' ) {
		if ( strtouint(str[2], &ver) )
			return 0;
		flags|=RPC_CHECK_VER;
	}

	if ( !(p=calloc(1, sizeof(*p))) )
		return 0;

	p->app = htonl(app);
	p->ver = htonl(ver);
	p->proc = htonl(proc);
	p->flags = flags;

	*priv = p;

	return 1;
}

static struct matcher rpc_matcher={
	.validate = rpc_validate,
	.free_priv = free,
	.new_node = rpc_new,
	.insert_node = rpc_insert,
};

static struct field builtin_fields[] = {
	{
		.name = "rpc",
		.desc = "Remote Procedure Call",
		.cost = MCOST_APP + 12,
		.type = IDX_INVALID,
		.matcher = &rpc_matcher,
	},
	{.name = NULL }
};

struct field_reg firestorm_fields[] = {
	{ .field = builtin_fields },
	{ .field = NULL }
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "decode.rpc",
	.desc		= "RPC matching routines",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 2,
	.ver_minor	= 0,
};
