/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/
#include "tcpip.h"

/* Generator */
struct generator tcp_gen=init_generator("sig.tcp", NULL);

static int  tcp_serialize(struct packet *, int, char *, size_t);
static void *tcp_deserialize(struct packet*, int, const char *, size_t);

static struct sub_table tcp_subs[]={
	{"src", "ip.src"},
	{"dst", "ip.dst"},
	{"tos", "ip.tos"},
	{"id", "ip.id"},
	{"fragoffset", "ip.fragofs"},
	{"fragbits", "ip.fragbits"},
	{"ttl", "ip.ttl"},
	{"ip_proto", "ip.proto"},
	{"sameip", "ip.sameip"},
	{"ipopts", "ip.opts"},
	{"sport", "tcp.sport"},
	{"dport", "tcp.dport"},
	{"seq", "tcp.seq"},
	{"ack", "tcp.ack"},
	{"win", "tcp.win"},
	{"window", "tcp.win"},
	{"statelss", "tcp.stateless"},
	{"flow", "tcp.flow"},
	{"flags", "tcp.flags"},
	{"http_method", "http.method"},
	{NULL, NULL},
};

struct proto tcp_p={
	.name		= "tcp",
	.decode		= tcp_decode,
	.print		= tcp_dprint,
	.sig_add	= tcp_commit,
	.sig_match	= tcp_match,
	.sig_subs	= tcp_subs,
	.serialize	= tcp_serialize,
	.deserialize	= tcp_deserialize,
	.free		= free,
	.fields		= tcp_fields,
	.tag		= tcp_tag,
};
struct proto_req tcp_r[]={
	proto_request("ip", 6),
	null_request()
};

static void tag_tuple(struct alert_tag *t, u_int32_t saddr, u_int32_t daddr,
			u_int16_t spt, u_int16_t dpt)
{
}

void tcp_tag(struct packet *p, int l, struct alert_tag *t)
{
	struct pkt_iphdr *iph = p->layer[l-1].h.ip;
	struct pkt_tcphdr *tcph = p->layer[l].h.tcp;

	if ( tcp_stateful && t->type == TAG_TYPE_SESSION ) {
		tcpstream_tag_session(p, l, t);
		return;
	}

	tag_tuple(t, iph->saddr, iph->daddr, tcph->sport, tcph->dport);
}

/* for serialize/deserialize */
/* XXX: are htonl() and ntohl() interchangable on all CPUs? */
static void tcp_swapbytes(struct tcp_stream *s, struct tcp_stream *f)
{
	f->state=s->state;
	f->flags=s->flags;
	f->scale=s->scale;
	f->queue=s->queue;
	f->ts_recent=htonl(s->ts_recent);
	f->ts_recent_stamp=htonl(s->ts_recent_stamp);
	f->snd_una=htonl(s->snd_una);
	f->snd_nxt=htonl(s->snd_nxt);
	f->rcv_nxt=htonl(s->rcv_nxt);
	f->rcv_wnd=htonl(s->rcv_wnd);
	f->rcv_wup=htonl(s->rcv_wup);
	f->isn=htonl(s->isn);
}

static int tcp_serialize(struct packet *p, int i, char *buf, size_t len)
{
	struct tcp_session *s=(struct tcp_session *)p->layer[i].session;
	struct tcp_serial  *f=(struct tcp_serial  *)buf;

	if ( len < sizeof(struct tcp_serial) )
		return -1;

	/* Already network byte order */
	f->c_addr=s->c_addr;
	f->s_addr=s->s_addr;
	f->c_port=s->c_port;
	f->s_port=s->s_port;

	tcp_swapbytes(&s->client, &f->client);
	tcp_swapbytes(&s->server, &f->server);

	return sizeof(*f);
}

/* Remember: this data is un-trusted so validate it! */
static void *tcp_deserialize(struct packet *p, int i,
				const char *buf, size_t len)
{
	struct tcp_session *s;
	struct tcp_serial *f=(struct tcp_serial *)buf;

	if ( len != sizeof(*f) )
		return NULL;

	/* validate input */
	if ( f->server.state >= TCP_MAX_STATES )
		return NULL;
	if ( f->client.state >= TCP_MAX_STATES )
		return NULL;

	s = calloc(1, sizeof(*s));
	if ( s == NULL )
		return NULL;

	/* already network byte order */
	s->c_addr=f->c_addr;
	s->s_addr=f->s_addr;
	s->c_port=f->c_port;
	s->s_port=f->s_port;

	tcp_swapbytes(&f->client, &s->client);
	tcp_swapbytes(&f->server, &s->server);

	return s;
}

int tcp_dprint(struct packet *p, int l, char *buf, size_t buflen)
{
	struct layer *a = &p->layer[l];
	const char tcpflags[]="FSRPAU12";
	char disp[9];
	int n, x;

	for(n=0,x=1; n<8; n++, x*=2) {
		if ( a->h.tcp->flags.flags & x ){
			disp[n]=tcpflags[n];
		}else{
			disp[n]='*';
		}
	}disp[8]='\0';

	return snprintf(buf, buflen,
		"%u > %u [%s] seq=0x%x ack=0x%x win=%u",
		ntohs(a->h.tcp->sport),
		ntohs(a->h.tcp->dport),
		disp,
		ntohl(a->h.tcp->seq),
		ntohl(a->h.tcp->ack),
		ntohs(a->h.tcp->win));
}

static inline u_int32_t tcp_csum(struct packet *pkt, int l)
{
#ifdef TCP_ASSUME_CSUM_OK
	return FLAG_TCP_CSUM;
#else
	struct pkt_iphdr *iph=pkt->layer[l].h.ip;
	struct pkt_tcphdr *tcph=pkt->layer[l+1].h.tcp;
	struct tcp_phdr ph;
	u_int16_t *tmp;
	u_int32_t sum=0;
	u_int16_t len, csum;
	int i;

	if ( pkt->layer[l].flags&FLAG_IP_DTRUNC )
		return 0;

	len=ntohs(iph->tot_len) - (iph->ihl<<2);

	/* Make pseudo-header */
	ph.sip=iph->saddr;
	ph.dip=iph->daddr;
	ph.zero=0;
	ph.proto=iph->protocol;
	ph.tcp_len=htons(len);

	/* Checksum the pseudo-header */
	tmp=(u_int16_t *)&ph;
	for(i=0; i<6; i++) {
		sum+=tmp[i];
	}

	/* Checksum the header+data */
	tmp=(u_int16_t *)tcph;
	for(i=0; i<len>>1; i++) {
		sum+=tmp[i];
	}

	/* Deal with last byte (if odd number of bytes) */
	if ( len&1 ) {
		union {
			u_int8_t b[2];
			u_int16_t s;
		}f;

		f.b[0]=((u_int8_t *)tcph)[len-1];
		f.b[1]=0;
		sum+=f.s;
	}

	sum = (sum & 0xffff) + (sum >> 16);

	csum = (~sum) & 0xffff;

	return csum ? 0 : FLAG_TCP_CSUM;
#endif
}

void tcp_decode(struct packet *p)
{
	struct layer *l=&p->layer[p->llen];
	size_t hlen;

	/* Check it is at least minimum size */
	if ( l->h.raw+sizeof(struct pkt_tcphdr) > p->end ) {
		goto finish;
	}

	/* Grab the header length */
	hlen = l->h.tcp->doff << 2;
	if ( hlen < sizeof(struct pkt_tcphdr) )
		goto finish;

	/* Use the doff field to decode the packet */
	p->layer[p->llen+1].h.raw = l->h.raw + hlen;
	if ( p->layer[p->llen+1].h.raw > p->end ) {
		goto finish;
	}

	p->llen++;

	if ( p->llen >= PKT_LAYERS ) {
		dispatch(p);
		return;
	}

	/* State tracking */
	if ( tcp_stateful ) {
		l->flags|=tcp_csum(p, p->llen-2);
		tcpstream_process(p, p->llen-2);
	}else{
finish:
		if ( p->layer[p->llen].h.raw < p->end )
			p->layer[p->llen++].proto=NULL;
		dispatch(p);
	}
}


/* ===================================================
 * PACKET MATCHING STUFF BEYOND THIS POINT
 * ===================================================
 */

/* do_stateful:
 * If tcpstream is enabled, and there is no flow/stateless matcher
 * then we insert the default one, to make sure all rules get
 * stateful inspection
 */
static int do_stateful(struct rule *r)
{
	struct criteria *c;

	for(c=r->criteria; c; c=c->next) {
		if ( strcmp(c->match->owner->name, "tcp.stateless") ||
			strcmp(c->match->owner->name, "tcp.flow") ) {
			return 1;
		}
	}

	return signature_criteria(r, 0, "flow", NULL);
}

static struct detect_root tcp_root;

void tcp_match(struct packet *p, int l)
{
	detect(&tcp_root, p, --l);
}

static const char * const stateful_stuff[]={
	"flow",
	"stateless",
	NULL,
};

int tcp_commit(struct rule *r)
{
	struct alert *a;
	struct sig_node *x;
	const char **ignore = NULL;

	if ( tcp_stateful ) {
		if ( !do_stateful(r) )
			return 0;
	}else{
		ignore = (const char **)stateful_stuff;
	}

	x = detect_new_sig(r, ignore);
	if ( x == NULL )
		return 0;

	a = calloc(1, sizeof(*a));
	if ( a == NULL ) {
		detect_free_sig(x, r->num_criteria);
		return 0;
	}

	memcpy(a, &r->alert, sizeof(*a));
	a->gen = &tcp_gen;

	return detect_add_sig(&tcp_root, x, r->num_criteria, a);
}
