/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
 *
 * tcpstream - Firestorm TCP state tracking implementation.
 *
 * TODO:
 *  o LRU stream reassembly, reap on high watermark (do not timeout)
 *  o Make timeout list a priority queue
 *  o Implement timeouts for SYN2/FIN states
 *  o Implement TIME_WAIT states
 *  o Configurable timeouts
 *  o Re-order out-of-order transmissions
 *  o Support SACK
 *  o Optimizations
 *  o Connection re-sync ?
 *  o Connection pickup ?
 *  o Better allocator for tcp stream reassembly?
 *
 * Should CORRECTLY cope with:
 *  o Re-transmissions
 *  o Bad TCP or IP checksum evasion
 *  o Broadcast/Multicast packets
 *  o PAWS (rfc1323) evasion
 *  o In-order stream reassembly
 *
 * Should ideally cope with:
 *  o Out-of-order stream reassembly
 *  o TIME_WAIT states
 *  o Packet loss
*/

#ifndef DONT_INLINE
#define __INLINE__ static inline
#else
#define __INLINE__ static
#endif

#ifndef STATE_DEBUG
#define dmesg(x...)
#else
#define dmesg mesg
static char *state_str[]={
	"CLOSED",
	"ESTABLISHED",
	"SYN_SENT",
	"SYN_RECV",
	"FIN_WAIT1",
	"FIN_WAIT2",
	"TCP_TIME_WAIT",
	"TCP_CLOSE",
	"TCP_CLOSE_WAIT",
	"TCP_LAST_ACK",
	"TCP_LISTEN",
	"TCP_CLOSING"
};
#endif

#include "tcpip.h"

#define TCP_PAWS_24DAYS (60 * 60 * 24 * 24)
#define TCP_PAWS_MSL 60
#define TCP_PAWS_WINDOW 60

#define TCP_TMO_SYN1 (90 * TIMESTAMP_HZ)

/* Minimum and maximum sizes for stream reassembly buffers */
#define TCP_BUFFER_GRANULE_BITS 7
#define TCP_BUFFER_GRANULE (1<<TCP_BUFFER_GRANULE_BITS)
static size_t tcp_maxstream=TCP_BUFFER_GRANULE * 32;
static size_t tcp_stream_acct = 32768;

/* Configuration */
static unsigned int tcp_minttl;
static unsigned int tcp_numstreams;
static unsigned int tcp_numflows;
static unsigned int tcp_reassemble;
unsigned int tcp_stateful = 0;

static struct arg args[]={
	{"minttl", ARGTYPE_PUINT, {vp_uint:&tcp_minttl}, "1"},
	{"num_streams", ARGTYPE_PBYTES, {vp_bytes:&tcp_numstreams}, "2048"},
	{"num_flows", ARGTYPE_PBYTES, {vp_bytes:&tcp_numflows}, "1024"},
	{"reassemble", ARGTYPE_PBOOL, {vp_bool:&tcp_reassemble}, "true"},
	{NULL,}
};

/* TCP Alerts */
static struct tokenbucket tcpstream_tb={
	.cost=(RATE_SEC/1), /* 1 alert per second */
	.burst=(RATE_SEC/1)*10, /* a burst of 10 alerts per second */
};

static struct generator tcpstream_gen=
	init_generator("tcpstream", &tcpstream_tb);

static struct alert alert_tcp1={
	.alert="In-window SYN",
	.sid=1, .rev=0,
	.priority=5,
	.gen=&tcpstream_gen,
	.order = ALERT_BUILTIN,
};
static struct alert alert_tcp2={
	.alert="Data sent on closed stream",
	.sid=2, .rev=0,
	.priority=5,
	.gen=&tcpstream_gen,
	.order = ALERT_BUILTIN,
};
static struct alert alert_tcp3={
	.alert="PAWS discard",
	.sid=3, .rev=0,
	.priority=5,
	.gen=&tcpstream_gen,
	.order = ALERT_BUILTIN,
};
static struct alert alert_tcp4={
	.alert="Reassembly Error",
	.sid=4, .rev=0,
	.priority=5,
	.gen=&tcpstream_gen,
	.order = ALERT_BUILTIN,
};
static struct alert alert_tcp5={
	.alert="Malicious TCP options",
	.sid=5, .rev=0,
	.priority=5,
	.gen=&tcpstream_gen,
	.order = ALERT_BUILTIN,
};

/* ICMP Alerts */
static struct generator icmperr_gen=
	init_generator("tcp.icmperr", &tcpstream_tb);

static struct alert alert_icmp1={
	.alert="ICMP error from interloper",
	.sid=1, .rev=0,
	.priority=5,
	.gen=&icmperr_gen,
	.order = ALERT_BUILTIN,
};
static struct alert alert_icmp2={
	.alert="ICMP error on established connection",
	.sid=2, .rev=0,
	.priority=5,
	.gen=&icmperr_gen,
	.order = ALERT_BUILTIN,
};

/* 1. Hash table for lookups -- O(n) in pathological cases, can be O(1) */
static unsigned int tcp_hashsz=512;
static struct tcp_session **tcp_hash;

/* 2. Slab cache for allocation -- O(1) */
static union tcp_union *all_sessions;
static union tcp_union *tcp_next;

/* 3. LRU list to evict oldest items -- O(1) */
static struct tcp_lru lru={
	.next=(struct tcp_session *)&lru,
	.prev=(struct tcp_session *)&lru,
};

/* 4. Timeout list to evict entries on timeout */
static struct tcp_tmo syn1={
	.next=(struct tcp_session *)&syn1,
	.prev=(struct tcp_session *)&syn1,
};

/* O(1) allocator for flows */
static size_t flow_len;
static void *flow_cache;
static void *flow_next;

/* The current packet */
static struct {
	struct tcpseg tp;
	struct pkt_iphdr *iph;
	struct pkt_tcphdr *tcph;
	struct tcp_session *s;
	struct tcp_stream *snd, *rcv;
	struct tcp_sbuf *sndbuf, *rcvbuf;
	char *ptr;
	int free;
	int layer;
	struct packet *pkt;
}cur;

/* Statistics counters */
static unsigned int tcp_packets;
static unsigned int tcp_state_errs;
static unsigned int tcp_reassembled;
static unsigned int tcp_broadcast;
static unsigned int tcp_lowttl;
static unsigned int tcp_timeouts;
static unsigned int max_concurrent;
static unsigned int num_active;
static unsigned int max_flows;
static unsigned int num_flows;

/* Wrap-safe seq/ack calculations */
__INLINE__ int between(u_int32_t s1, u_int32_t s2, u_int32_t s3) {
	return s3 - s2 >= s1 - s2; /* is s2<=s1<=s3 ? */
}
__INLINE__ u_int32_t tcp_receive_window(struct tcp_stream *tp)
{
	int32_t win=tp->rcv_wup+tp->rcv_wnd-tp->rcv_nxt;
	if ( win<0 )
		win=0;
	return (u_int32_t)win;
}
__INLINE__ int tcp_sequence(struct tcp_stream *tp, u_int32_t seq, u_int32_t end_seq)
{
	return !tcp_before(end_seq, tp->rcv_wup) &&
		!tcp_after(seq, tp->rcv_nxt+tcp_receive_window(tp));
}

/* Frees up the reassembly buffer for a stream */
__INLINE__ void tcpstream_freebuf(struct tcp_stream *s, struct tcp_sbuf *b)
{
	if ( b->buf ) {
		free(b->buf);
		tcp_stream_acct += b->buf_sz;
	}
	b->buf=NULL;
	b->buf_sz=0;
	b->buf_ofs=0;
	s->queue=0;
}

/* Grow a TCP stream reassembly buffer */
static void tcp_buffer_resize(struct tcp_sbuf *b, size_t len)
{
	char *new;
	size_t ofs;

	/* Round up to our granularity */
	len = (len + (TCP_BUFFER_GRANULE-1)) & ~(TCP_BUFFER_GRANULE-1);
	if ( len <= b->buf_sz )
		return;

	/* Don't allocate more than is allowed */
	ofs = len - b->buf_sz;
	if ( tcp_stream_acct < ofs ) {
		ofs = tcp_stream_acct;
		len = b->buf_sz + tcp_stream_acct;
	}

	if ( len == 0 )
		return;

	new = realloc(b->buf, len);
	if ( new == NULL ) {
		mesg(M_DEBUG, "tcpstream: %u -> %u (%u): realloc: %s",
			b->buf_sz, len, tcp_stream_acct, sys_err());
		return;
	}

	tcp_stream_acct -= ofs;

	dmesg(M_DEBUG, "tcpstream: acct: %u -> %u (%u)",
		b->buf_sz, len, tcp_stream_acct);

	b->buf = new;
	b->buf_sz = len;
}

/* TMO: Add a TCP session to the head of a timeout list */
__INLINE__ void tcp_tmo_add(struct tcp_tmo *l, struct tcp_session *s)
{
	s->tmo_next=l->next;
	s->tmo_prev=(struct tcp_session *)l;
	l->next->tmo_prev=s;
	l->next=s;
}

/* TMO: Remove a TCP session from a timeout list */
__INLINE__ void tcp_tmo_del(struct tcp_session *s)
{
	if ( likely(s->expire == 0) )
		return;

	s->tmo_prev->tmo_next=s->tmo_next;
	s->tmo_next->tmo_prev=s->tmo_prev;
	s->tmo_prev=NULL;
	s->tmo_next=NULL;
	s->expire=0;
}

/* TMO: Check timeouts */
__INLINE__ void tcp_tmo_check(struct packet *pkt)
{
	timestamp_t now = pkt->time;

	while ( syn1.prev != (struct tcp_session *)&syn1 ) {
		if ( !time_before(syn1.prev->expire, now) )
			return;
		tcp_free(syn1.prev);
		tcp_timeouts++;
	}
}

/* TMO: Set expiry */
__INLINE__ void tcp_expire(struct packet *p,
				struct tcp_session *s,
				timestamp_t t)
{
	tcp_tmo_del(s);
	s->expire = p->time + t;
	tcp_tmo_add(&syn1, s);
}

/* LRU: Add a TCP session to the head of an LRU list */
__INLINE__ void tcp_lru_add(struct tcp_lru *l, struct tcp_session *s)
{
	s->next=l->next;
	s->prev=(struct tcp_session *)l;
	l->next->prev=s;
	l->next=s;
}

/* LRU: Remove a TCP session from an LRU */
__INLINE__ void tcp_lru_del(struct tcp_session *s)
{
	s->prev->next=s->next;
	s->next->prev=s->prev;
}

/* LRU: Move a TCP session to the front of an LRU */
__INLINE__ void tcp_lru_mtf(struct tcp_lru *l, struct tcp_session *s)
{
	tcp_lru_del(s);
	tcp_lru_add(l,s);
}

/* HASH: Unlink a session from the session hash */
__INLINE__ void tcp_hash_unlink(struct tcp_session *s)
{
	if ( s->hash_next )
		s->hash_next->hash_pprev=s->hash_pprev;
	*s->hash_pprev=s->hash_next;
}

/* HASH: Link a session in to the TCP session hash */
__INLINE__ void tcp_hash_link(struct tcp_session *s)
{
	if ( (s->hash_next=tcp_hash[s->bucket]) ) {
		s->hash_next->hash_pprev=&s->hash_next;
	}
	tcp_hash[s->bucket]=s;
	s->hash_pprev=&tcp_hash[s->bucket];
}

/* HASH: Move to front of hash collision chain */
__INLINE__ void tcp_hash_mtf(struct tcp_session *s)
{
	tcp_hash_unlink(s);
	tcp_hash_link(s);
}

/* Flow cache allocation, slab-cache stylee */
/* TODO: strategy for when maximum is reached, evict LRU? */
__INLINE__ void *tcp_flow_alloc(void)
{
	void *ret=flow_next;
	if ( !ret ) return ret;
	flow_next=*(void **)flow_next;
	if ( ++num_flows > max_flows )
		max_flows=num_flows;
	return ret;
}

__INLINE__ void tcp_flow_free(void *p)
{
	*(void **)p=flow_next;
	flow_next=p;
	num_flows--;
}

static void do_tag(struct tcp_tag *t, struct packet *p, struct alert_tag *tag)
{
	t->metric = tag->metric;
	switch ( t->metric ) {
	case TAG_METRIC_PACKETS:
	case TAG_METRIC_BYTES:
		t->count = tag->count;
		break;
	case TAG_METRIC_SECONDS:
		t->count = p->time + (tag->count * TIMESTAMP_HZ);
		break;
	}
	t->alert = p->alert;
}

void tcpstream_tag_session(struct packet *p, int l, struct alert_tag *t)
{
	struct tcp_session *s = p->layer[l].session;
	struct pkt_iphdr *iph = p->layer[l-1].h.ip;
	struct pkt_tcphdr *tcph = p->layer[l].h.tcp;

	switch ( t->direction ) {
	case TAG_DIR_SERVER:
		do_tag(&s->s_tag, p, t);
		break;
	case TAG_DIR_CLIENT:
		do_tag(&s->c_tag, p, t);
		break;
	case TAG_DIR_SRC_ONLY:
		if ( (iph->saddr == s->s_addr) &&
			(tcph->sport == s->s_port)) {
			do_tag(&s->s_tag, p, t);
		}else{
			do_tag(&s->c_tag, p, t);
		}
		break;
	case TAG_DIR_DST_ONLY:
		if ( (iph->daddr == s->s_addr) &&
			(tcph->dport == s->s_port) ) {
			do_tag(&s->s_tag, p, t);
		}else{
			do_tag(&s->c_tag, p, t);
		}
		break;
	case TAG_DIR_BOTH:
		do_tag(&s->c_tag, p, t);
		do_tag(&s->s_tag, p, t);
		break;
	default:
		mesg(M_CRIT, "tcpstream: tag had fucked up direction: %u",
			t->direction);
		return;
	}

	mesg(M_DEBUG, "sv=%.8x:%u cl=%.8x:%u s_tag=%s c_tag=%s",
		s->s_addr, s->s_port,
		s->c_addr, s->c_port,
		s->s_tag.count ? "yes" : "no",
		s->c_tag.count ? "yes" : "no");

}

/* Parse TCP options just for timestamps */
static int tcp_fast_options(struct pkt_tcphdr *t, u_int32_t *tsval)
{
	char *tmp, *end;
	size_t ofs = t->doff << 2;

	/* Return if we don't have any */
	if ( ofs <= sizeof(struct pkt_tcphdr))
		return 0;

	/* Work out where they begin and end */
	tmp=end=(char *)t;
	tmp+=sizeof(struct pkt_tcphdr);
	end+=ofs;

	while ( tmp<end ) {
		size_t step;

		/* XXX: We continue past an EOL. Is that right? */
		switch(*tmp) {
		case TCPOPT_EOL:
		case TCPOPT_NOP:
			tmp++;
			continue;
		}

		if ( tmp+1 >= end )
			break;

		switch(*tmp) {
		case TCPOPT_TIMESTAMP:
			if ( tmp+10 >= end )
				break;
			*tsval=ntohl(*((u_int32_t *)(tmp+2)));
			return 1;
		}

		step = *(tmp+1);
		if ( step < 2 ) {
			alert_tag(cur.pkt, &alert_tcp5, -1);
			step = 2;
		}
		tmp += step;
	}

	return 0;
}

/* This will parse TCP options for SYN packets */
static void tcp_syn_options(struct tcp_stream *s, struct pkt_tcphdr *t, u_int32_t sec)
{
	char *tmp, *end;
	size_t ofs = t->doff << 2;

	/* Return if we don't have any */
	if ( ofs <= sizeof(struct pkt_tcphdr))
		return;

	/* Work out where they begin and end */
	tmp=end=(char *)t;
	tmp+=sizeof(struct pkt_tcphdr);
	end+=ofs;

	while ( tmp<end ) {
		size_t step;

		/* XXX: We continue past an EOL. Is that right? */
		switch(*tmp) {
		case TCPOPT_EOL:
		case TCPOPT_NOP:
			tmp++;
			continue;
		}

		if ( tmp+1 >= end )
			break;

		/* Deal with fixed size options */
		switch(*tmp) {
		case TCPOPT_SACK_PERMITTED:
			s->flags |= TF_SACK_OK;
			break;
		case TCPOPT_TIMESTAMP:
			s->flags |= TF_TSTAMP_OK;

			/* Only check the bit we want */
			if ( tmp+10 >= end )
				break;

			s->ts_recent=ntohl(*((u_int32_t *)(tmp+2)));
			s->ts_recent_stamp=sec;

			break;
		case TCPOPT_WSCALE:
			if ( tmp+2 >= end )
				break;

			s->flags|=TF_WSCALE_OK;

			/* rfc1323: must log error and limit to 14 */
			if ( (s->scale=*(tmp+2)) > 14 )
				s->scale=14;
			break;
		}

		step = *(tmp+1);
		if ( step < 2 ) {
			alert_tag(cur.pkt, &alert_tcp5, -1);
			step = 2;
		}
		tmp += step;
	}
}

/* Perform a state transition */
__INLINE__ void transition(struct tcp_session *t, int c, int s)
{
#ifdef STATE_DEBUG
	char cip[32],sip[32];
	struct in_addr ca, sa;

	ca.s_addr=t->c_addr;
	sa.s_addr=t->s_addr;
	snprintf(cip, sizeof(cip), "%s", inet_ntoa(ca));
	snprintf(sip, sizeof(sip), "%s", inet_ntoa(sa));

	dmesg(M_DEBUG, "tcpstream: %s:%u(%s) -> %s:%u(%s)",
		cip, ntohs(t->c_port), state_str[c],
		sip, ntohs(t->s_port), state_str[s]);
#endif
	t->client.state=c;
	t->server.state=s;
}

/* Hash function. Hashes to the same value even
 * when source and destinations are inverted */
__INLINE__ int tcp_hashfn(u_int32_t saddr, u_int32_t daddr,
	u_int16_t sport, u_int16_t dport)
{
	int h = ((saddr ^ sport) ^ (daddr ^ dport));
	h ^= h>>16;
	h ^= h>>8;
	return h % tcp_hashsz;
}

/* Return a TCP session descriptor to the cache */
void tcp_free(struct tcp_session *s)
{
	/* get the debug message */
	transition(s, 0, 0);

	/* HASH: unhash */
	tcp_hash_unlink(s);

	/* LRU: remove */
	tcp_lru_del(s);

	/* TMO: remove */
	tcp_tmo_del(s);

	if ( s->flow ) {
		if ( s->proto && s->proto->flow_dtor )
			s->proto->flow_dtor(s->flow);
		tcp_flow_free(s->flow);
	}

	tcpstream_freebuf(&s->client, &s->c_buf);
	tcpstream_freebuf(&s->server, &s->s_buf);

	/* CACHE: free */
	((union tcp_union *)s)->next=tcp_next;
	tcp_next=(union tcp_union *)s;

	num_active--;
}

/* Add a new session, never fails */
static struct tcp_session *tcp_new(struct pkt_iphdr *iph,
	struct pkt_tcphdr *tcph, struct packet *p)
{
	struct tcp_session *ret;

	/* If we have run out of connections, then
	 * just free the oldest one in the LRU */
	if ( !tcp_next ) {
		tcp_free(lru.prev);
		tcp_timeouts++;
	}

	/* CACHE: alloc */
	ret=&tcp_next->s;
	tcp_next=tcp_next->next;

	/* Keep track of max number of concurrent sessions
	 * seen so far */
	if ( ++num_active > max_concurrent )
	       max_concurrent=num_active;

	/* Zero values */
	memset(&ret->client, 0, sizeof(ret->client));
	memset(&ret->server, 0, sizeof(ret->server));

	/* Setup the key values */
	ret->c_addr=iph->saddr;
	ret->s_addr=iph->daddr;
	ret->c_port=tcph->sport;
	ret->s_port=tcph->dport;
	ret->bucket=tcp_hashfn(ret->c_addr, ret->s_addr,
		ret->c_port, ret->s_port);

	/* Setup initial state */
	transition(ret, TCP_SYN_SENT, 0);
	ret->server.isn=ntohl(tcph->seq);
	ret->client.snd_una=ret->server.isn+1;
	ret->client.snd_nxt=ret->client.snd_una+1;
	ret->client.rcv_wnd=ntohs(tcph->win);

	ret->server.rcv_nxt=ret->client.snd_una;
	ret->server.rcv_wup=ret->client.snd_una;

	/* server sees clients initial options */
	tcp_syn_options(&ret->server, tcph, p->time / TIMESTAMP_HZ);

	if ( (ret->proto=decode_next(&tcp_p, ret->s_port))
		&& ret->proto->flow_len ) {
		ret->flow=tcp_flow_alloc();
		if ( ret->flow && ret->proto->flow_ctor )
			ret->proto->flow_ctor(ret->flow);
	}else{
		/* TODO: auto-detect protocol ;) */
		ret->flow=NULL;
	}

	/* Set the timeout */
	tcp_expire(p, ret, TCP_TMO_SYN1);

	tcp_lru_add(&lru, ret);
	tcp_hash_link(ret);

	return ret;
}

/* Find a TCP session given a packet */
static struct tcp_session *tcp_find(struct pkt_iphdr *iph,
	struct pkt_tcphdr *tcph,
	int *to_server)
{
	struct tcp_session *s;
	int idx=tcp_hashfn(iph->saddr, iph->daddr,
		tcph->sport, tcph->dport);

	for(s=tcp_hash[idx]; s; s=s->hash_next) {
		if (	s->s_addr==iph->saddr &&
			s->c_addr==iph->daddr &&
			s->s_port==tcph->sport &&
			s->c_port==tcph->dport ) {
			*to_server=0;
			return s;
		}
		if (	s->c_addr==iph->saddr &&
			s->s_addr==iph->daddr &&
			s->c_port==tcph->sport &&
			s->s_port==tcph->dport ) {
			*to_server=1;
			return s;
		}
	}

	return NULL;
}

/* rfc793: Actions to perform when recieving an ACK in
 * an established state */
__INLINE__ void tcp_established(struct tcp_stream *snd,
	struct tcp_stream *rcv,
	u_int32_t seq,
	u_int32_t ack,
	u_int32_t win)
{
	if ( tcp_after(ack, rcv->snd_una) && !tcp_before(ack, rcv->snd_nxt) )
		rcv->snd_una=ack;

	snd->snd_una=seq;
	snd->snd_nxt=seq+1;

	snd->rcv_wup=ack;
	snd->rcv_wnd=win;
}

/* A segment contained data so check its validity */
static void tcpstream_data(void)
{
	char *ptr=(char *)cur.tcph + (cur.tcph->doff<<2);
	struct tcpseg *tp=&cur.tp;
	struct tcp_stream *rcv=cur.rcv;

	if ( tp->seq==rcv->rcv_nxt ) {
		if (tcp_receive_window(rcv) == 0) {
			/* XXX: Alert here? */
			return;
		}

		cur.ptr = ptr;
		rcv->rcv_nxt = tp->seq_end;
	}else{
		/* A retransmit */
		if ( !tcp_after(tp->seq_end,rcv->rcv_nxt) )
			return;

		/* TODO: Always need to re-order any out of order
		 * transmissions. Not sure what kind of mem policy
		 * we are going to need here...
		 */
		dmesg(M_DEBUG, "tcpstream: Out of order: %u->%u (wanted %u)",
			tp->seq-rcv->isn,
			tp->seq_end-rcv->isn,
			rcv->rcv_nxt-rcv->isn);
	}
}

/* Process incoming TCP segments here */
static void tcpstream_tcpseg(unsigned int i)
{
	struct tcp_session *s;
	struct tcp_stream *snd, *rcv;
	struct pkt_tcphdr *tcph=cur.tcph;
	struct pkt_iphdr *iph=cur.iph;
	struct tcpseg *tp=&cur.tp;
	struct packet *p=cur.pkt;
	struct tcp_tag *snd_tag;
	int to_server;

	/* Find the session this packet comes from */
	if ( !(s=tcp_find(iph, tcph, &to_server)) ) {
		if ( (tcph->flags.bits.syn) &&
			!(tcph->flags.bits.ack) &&
			!(tcph->flags.bits.rst)) {

			/* SYN: part 1 of connection handshake */
			p->layer[i+1].session=tcp_new(iph, tcph, p);
			p->layer[i+1].flags|=FLAG_TCP_STATE|FLAG_TCP_2SVR;
			return;
		}

		dmesg(M_DEBUG,"tcpstream: never heard of you");
		return;
	}else p->layer[i+1].session=s;

	/* figure out who is sender and who is reciever */
	cur.s = s;

	if ( to_server ) {
		snd=cur.snd = &s->client;
		rcv=cur.rcv = &s->server;
		snd_tag = &s->c_tag;
		cur.sndbuf = &s->c_buf;
		cur.rcvbuf = &s->s_buf;
		p->layer[i+1].flags |= FLAG_TCP_2SVR;
	}else{
		snd=cur.snd = &s->server;
		rcv=cur.rcv = &s->client;
		snd_tag = &s->s_tag;
		cur.sndbuf = &s->s_buf;
		cur.rcvbuf = &s->c_buf;
		p->layer[i+1].flags &= (~FLAG_TCP_2SVR);
	}

	if ( snd_tag->metric != TAG_METRIC_NONE ) {
		mesg(M_DEBUG, "Got a tagged packet");

		switch ( snd_tag->metric ) {
		case TAG_METRIC_PACKETS:
			if ( --snd_tag->count == 0 ) {
				snd_tag->metric = TAG_METRIC_NONE;
			}
			break;
		case TAG_METRIC_SECONDS:
			if ( cur.pkt->time >= (snd_tag->count * TIMESTAMP_HZ) ) {
				snd_tag->metric = TAG_METRIC_NONE;
			}
			break;
		case TAG_METRIC_BYTES:
			if ( cur.tp.len >= snd_tag->count ) {
				snd_tag->metric = TAG_METRIC_NONE;
			}else{
				snd_tag->metric -= cur.tp.len;
			}
			break;
		}
	}

	/* LRU: Move to front of LRU and hash collision chain */
	tcp_lru_mtf(&lru, s);
	tcp_hash_mtf(s);

	/* Deal with a SYN/ACK */
	if ( rcv->state==TCP_SYN_SENT ) {
		/* fist check the ack field */
		if ( tcph->flags.bits.ack ) {
		 	/* if SND.UNA =< SEG.ACK =< SND.NXT
			 * then ACK is acceptable */
			if ( !(between(tp->ack, rcv->snd_una, rcv->snd_nxt)) ) {
				return;
			}

			p->layer[i+1].flags|=FLAG_TCP_STATE|FLAG_TCP_SURE;
		}

		/* then check the rst flag */
		if ( tcph->flags.bits.rst ) {
			cur.free = 1;
			return;
		}

		if ( tcph->flags.bits.syn ) {
			/* update the advertised window */
			snd->rcv_wnd=tp->win;

			/* update sequencing information */
			snd->snd_una=tp->seq+1;
			snd->snd_nxt=snd->snd_una+1;

			rcv->isn=tp->seq;
			rcv->rcv_nxt=tp->seq_end;
			rcv->rcv_wup=tp->seq_end;
			if ( tcph->flags.bits.ack )
				rcv->snd_una=tp->ack;

			/* client now sees servers initial options */
			tcp_syn_options(&s->client, tcph, p->time / TIMESTAMP_HZ);

			/* Check whether to use window scaling */
			if (!(rcv->flags&TF_WSCALE_OK && snd->flags&TF_WSCALE_OK)) {
				rcv->scale=0;
				snd->scale=0;
			}

			/* SYN|ACK: part 2 of connection handshake */
			rcv->state=TCP_SYN_RECV;
			snd->state=TCP_SYN_SENT;
			transition(s, s->client.state, s->server.state);
			tcp_tmo_del(s);
			return;
		}else goto no_checks;
	}

	/* First check the sequence number */
	if ( !tcp_sequence(rcv, tp->seq, tp->seq_end) ) {
		/* XXX: Don't alert here retransmits hit this */
		return;
	}

no_checks:

	/* rfc1323: H1. Apply PAWS checks first */
	if ( rcv->flags & TF_TSTAMP_OK &&
		(tp->saw_tstamp=tcp_fast_options(tcph, &tp->tsval)) ) {
		/* TODO: PAWS is buggy as fuck. We use the Linux PAWS check,
		 * I have no idea what other stacks do...
		 */
		if ( (int32_t)(rcv->ts_recent - tp->tsval) > TCP_PAWS_WINDOW &&
			(u_int32_t)(p->time/TIMESTAMP_HZ) < rcv->ts_recent_stamp
			+ TCP_PAWS_24DAYS ) {
			alert_tag(cur.pkt, &alert_tcp3, -1);
			return;
		}
	}

	/* Second, check the RST bit */
	if ( tcph->flags.bits.rst ) {
		p->layer[i+1].flags|=FLAG_TCP_STATE|FLAG_TCP_SURE;
		dmesg(M_DEBUG,"TCP stream was reset");
		cur.free = 1;
		return;
	}

	/* rfc1323: PAWS: update ts_recent */
	if ( tp->saw_tstamp && !tcp_after(tp->seq,rcv->rcv_wup) ) {
		if((int32_t)(tp->tsval - rcv->ts_recent) >= 0 ||
		   (u_int32_t)(p->time/TIMESTAMP_HZ) >= rcv->ts_recent_stamp
		   + TCP_PAWS_24DAYS) {
			rcv->ts_recent=tp->tsval;
			rcv->ts_recent_stamp=p->time / TIMESTAMP_HZ;
		}
	}

	/* Third, check security and precendece (pfft) */

	/* Fourth, check the SYN bit */
	if ( tcph->flags.bits.syn && !tcp_before(tp->seq,rcv->rcv_nxt) ) {

		/* If not a retransmit, then it's an in-window SYN */
		if ( tp->seq != rcv->isn)
			alert_tag(cur.pkt, &alert_tcp1, -1);

		cur.free = 1;
		return;
	}

	/* we now know that this packet is in-state */
	p->layer[i+1].flags|=FLAG_TCP_STATE|FLAG_TCP_SURE;

	/* Scale the window */
	tp->win<<=snd->scale;

	/* Fifth, check the ack field */
	if ( !tcph->flags.bits.ack )
		goto no_ack;

	switch(rcv->state) {
	case TCP_SYN_SENT:
		/* ACK: part 3 of connection handshake */
		if ( rcv->snd_una <= tp->ack && tp->ack <= rcv->snd_nxt ) {
			p->layer[i+1].flags|=FLAG_TCP_CT_EST;
			tcp_established(snd,rcv,tp->seq,tp->ack,tp->win);
			transition(s, TCP_ESTABLISHED, TCP_ESTABLISHED);
		}
		break;
	case TCP_ESTABLISHED:
		tcp_established(snd,rcv,tp->seq,tp->ack,tp->win);
		break;
	case TCP_FIN_WAIT1:
		tcp_established(snd,rcv,tp->seq,tp->ack,tp->win);
		rcv->state=TCP_FIN_WAIT2;
		snd->state=TCP_LAST_ACK;
		transition(s, s->client.state, s->server.state);
	case TCP_FIN_WAIT2:
		break;
	case TCP_CLOSE_WAIT:
		tcp_established(snd,rcv,tp->seq,tp->ack,tp->win);
		break;
	case TCP_CLOSING:
		tcp_established(snd,rcv,tp->seq,tp->ack,tp->win);
		rcv->state=TCP_TIME_WAIT;
		transition(s, s->client.state, s->server.state);
		break;
	case TCP_LAST_ACK:
		rcv->state=0;
		transition(s, s->client.state, s->server.state);
		break;
	default:
		break;
	}

no_ack:
	/* sixth check URG bit */
	if ( !tcph->flags.bits.urg )
		goto no_urg;

	/* XXX: What does processing urgent/OOB data
	 * actually buy us in terms of intrusion detection
	 * capability? Need to look in to this... */

no_urg:
	/* seventh process the segment text */
	if ( !tp->len || !s->proto )
		goto no_data;

	/* There is data in the segment */
	switch(rcv->state) {
	case TCP_CLOSE_WAIT:
	case TCP_CLOSING:
	case TCP_LAST_ACK:
	case TCP_TIME_WAIT:
		alert_tag(cur.pkt, &alert_tcp2, -1);
		goto no_data;
	}

	if ( tcp_reassemble )
		tcpstream_data();

no_data:
	/* eighth, check the FIN bit */
	if ( !tcph->flags.bits.fin )
		goto no_fin;

	rcv->rcv_nxt++;

	switch(rcv->state) {
	case TCP_SYN_RECV:
	case TCP_ESTABLISHED:
		rcv->state=TCP_CLOSE_WAIT;
		snd->state=TCP_FIN_WAIT1;
		transition(s, s->client.state, s->server.state);
		break;
	case TCP_FIN_WAIT1:
		/* if fin has been acked do time_wait else closing */
		rcv->state=TCP_TIME_WAIT;
		transition(s, s->client.state, s->server.state);
		break;
	case TCP_FIN_WAIT2:
		rcv->state=TCP_TIME_WAIT;
		transition(s, s->client.state, s->server.state);
		break;
	case TCP_CLOSE_WAIT:
	case TCP_CLOSING:
	case TCP_LAST_ACK:
	case TCP_TIME_WAIT:
		/* Remain in the same state */
		break;
	default:
		break;
	}

no_fin:
	/* XXX: need to implement time_wait timer */
	if ( (snd->state==TCP_TIME_WAIT && rcv->state==0) ||
		(rcv->state==TCP_TIME_WAIT && snd->state==0) ) {
		cur.free=1;
	}else if ( rcv->state==0 && snd->state==TCP_SYN_SENT ) {
		/* XXX: we reset the timeout on retransmissions */
		if ( (tcph->flags.bits.syn) &&
			!(tcph->flags.bits.ack) &&
			!(tcph->flags.bits.rst)) {
			tcp_expire(p, s, TCP_TMO_SYN1);
		}
	}
}

/* Process an ICMP error message during connection */
void icmperr_process(struct packet *p, int l)
{
	struct pkt_iphdr *outer_iph;
	struct pkt_icmphdr *icmph;
	struct pkt_iphdr *iph;
	struct pkt_tcphdr *tcph;
	struct tcp_session *s;
	struct tcp_stream *snd,*rcv;
	int to_server;

	/* Get all the headers we need */
	if ( l+2 >= p->llen )
		return;
	outer_iph=p->layer[l-1].h.ip;
	icmph=p->layer[l].h.icmp;
	iph=p->layer[l+1].h.ip;
	tcph=p->layer[l+2].h.tcp;

	/* Make sure the packet is cool */
	if ( iph->protocol!=6 ) return;
	if ( icmph->type != ICMP_DEST_UNREACH ) return;
	if ( icmph->code == ICMP_FRAG_NEEDED ) return;
	if ( icmph->code > NR_ICMP_UNREACH ) return;

	/* Check the IP packet came from the endpoint
	 * ( if necessary ) */
	if ( icmph->code==ICMP_PROT_UNREACH ||
		icmph->code==ICMP_PORT_UNREACH ) {
		if ( iph->saddr!=outer_iph->daddr ) {
			alert_tag(p, &alert_icmp1, -1);
			return;
		}
	}

	/* Lookup our session */
	if ( !(s=tcp_find(iph, tcph, &to_server)) ) {
		/* XXX: Alert here? Seems to happen a lot... */
		return;
	}

	/* Find out which side sent the offending packet */
	if ( iph->daddr==s->c_addr ) {
		snd=&s->server;
		rcv=&s->client;
	}else{
		snd=&s->client;
		rcv=&s->server;
	}

	/* Only reset opening connections */
	if ( snd->state!=TCP_SYN_SENT && snd->state!=TCP_SYN_RECV ) {
		alert_tag(p, &alert_icmp2, -1);
		return;
	}

	/* TODO: Check seq/ack numbers */
	dmesg(M_DEBUG,"TCP reset by ICMP from %.8x", outer_iph->saddr);

	tcp_free(s);
}

/* Create a new packet based on the reassembled contents of
 * many segments.
 */
static char *tcpstream_reassemble(struct packet *pkt, struct packet *new, size_t t)
{
	size_t len, ilen, tlen;
	char *buf;

	if ( t > cur.rcvbuf->buf_ofs )
		t = cur.rcvbuf->buf_ofs;

	ilen = cur.iph->ihl<<2;
	tlen = cur.tcph->doff<<2;
	len = ilen + tlen + t;
	buf = malloc(len);
	if ( !buf )
		return NULL;

	memcpy(buf, cur.iph, ilen);
	memcpy(buf+ilen, cur.tcph, tlen);
	memcpy(buf+ilen+tlen, cur.rcvbuf->buf, t);

	packet_reinject_prepare(new, pkt);
	new->len = len;
	new->caplen = len;
	new->base = buf;
	new->end = buf + len;
	new->llen = 2;

	new->layer[0].proto=&ipv4_p;
	new->layer[0].h.raw=buf;
	new->layer[0].flags=FLAG_IP_CSUM;
	new->layer[0].session=NULL;

	new->layer[1].proto=&tcp_p;
	new->layer[1].h.raw=buf+ilen;
	new->layer[1].flags=FLAG_TCP_CSUM | FLAG_TCP_STATE |
		FLAG_TCP_SURE | FLAG_TCP_TRACK | FLAG_TCP_STREAM;
	new->layer[1].session=cur.s;

	new->layer[2].h.raw = buf + ilen + tlen;
	new->layer[2].proto = cur.s->proto;
	new->layer[2].flags = 0;
	new->layer[2].session = NULL;

	if ( cur.rcv == &cur.s->server )
		new->layer[1].flags|=FLAG_TCP_2SVR;

	/* Hack up the datagram length */
	new->layer[0].h.ip->tot_len=ntohs(len);

	/* Re-calculate checksums */
	/* TODO: partial checksum */
	ipfrag_csum(new->layer[0].h.ip);
	/* TODO: tcp checksum */

	dmesg(M_DEBUG, "tcpstream: %u byte reassemblygram", len);
	return buf;
}

/* Inject a tcp reassemblygram */
static void tcpstream_inject(struct packet *pkt, size_t t)
{
	struct packet p;
	char *buf;

	buf = tcpstream_reassemble(pkt, &p, t);
	if ( buf == NULL )
		return;

	/* Send it down the decode stack */
	p.layer[2].proto->decode(&p);

	tcp_reassembled++;

	free(buf);
}

/* Reassembly alert */
static void tcpstream_reasm_err(struct packet *pkt)
{
	struct packet p;
	struct packet *the_pkt = NULL;
	char *buf = NULL;

	if ( cur.rcvbuf->buf ) {
		buf = tcpstream_reassemble(pkt, &p, cur.rcvbuf->buf_ofs);
		the_pkt = &p;
	}

	if ( buf == NULL )
		the_pkt = pkt;

	alert_tag(the_pkt, &alert_tcp4, -1);

	free(buf);
}

/* Inject the TCP packet without decoding it */
__INLINE__ void tcpstream_nodecode(struct packet *pkt)
{
	/* Add the data layer if we have some data */
	if ( cur.tp.len && pkt->llen < PKT_LAYERS )
		pkt->layer[pkt->llen++].proto=NULL;

	dispatch(pkt);
}

/* Inject a TCP packet in to the application layer decoder */
static void tcpstream_decode(void)
{
	if ( cur.pkt->llen >= PKT_LAYERS ) {
		tcpstream_nodecode(cur.pkt);
		return;
	}

	cur.pkt->layer[cur.pkt->llen].flags=0;
	cur.pkt->layer[cur.pkt->llen].session=NULL;
	cur.pkt->layer[cur.pkt->llen].proto=cur.s->proto;
	cur.s->proto->decode(cur.pkt);
}

/* Add this packet to the back of the reassembly buffer */
static void tcp_stream_queue(void)
{
	struct tcp_session *s=cur.s;
	size_t len, t;

	/* Figure out how big the buffer will be */
	len = cur.rcvbuf->buf_ofs + cur.tp.len;

	/* If its bigger than we have then grow it */
	if ( len > cur.rcvbuf->buf_sz )
		tcp_buffer_resize(cur.rcvbuf, len);

	/* If the buffer space is smaller than the packet
	 * then just copy the first bit of the packet
	 */
	len = cur.rcvbuf->buf_sz - cur.rcvbuf->buf_ofs;
	len = (len > cur.tp.len ) ? cur.tp.len : len;

	/* There is no space left at all in the buffer */
	if ( len == 0 ) {
		tcpstream_reasm_err(cur.pkt);
		tcpstream_freebuf(cur.rcv, cur.rcvbuf);
		return;
	}

	/* Copy in to the buffer */
	memcpy(cur.rcvbuf->buf + cur.rcvbuf->buf_ofs, cur.ptr, len);
	cur.rcvbuf->buf_ofs += len;
	dmesg(M_DEBUG, "tcpstream: queuing %u/%u",
		cur.rcvbuf->buf_ofs, cur.rcvbuf->buf_sz);

	/* Now try to shove as much of the data as possible up the
	 * decode stack.
	 */
again:
	t = s->proto->sdecode(cur.pkt, cur.rcvbuf->buf, cur.rcvbuf->buf_ofs);
	if ( t == 0 ) {
		/* We have tried and we have failed, right up to the
		 * largest possible buffer size */
		if ( cur.rcvbuf->buf_ofs == tcp_maxstream ) {
			tcpstream_reasm_err(cur.pkt);
			tcpstream_freebuf(cur.rcv, cur.rcvbuf);
		}
		return;
	}

	/* Yay, reassembly has worked! */
	tcpstream_inject(cur.pkt, t);

	/* All of buffer was taken OK */
	if ( t >= cur.rcvbuf->buf_sz ) {
		tcpstream_freebuf(cur.rcv, cur.rcvbuf);
		return;
	}

	/* Some data remains so shift buffer down and
	 * try again
	 */
	cur.rcvbuf->buf_ofs -= t;
	memmove(cur.rcvbuf->buf, cur.rcvbuf->buf + t, cur.rcvbuf->buf_ofs);
	goto again;
}

/* Perform packet queuing for TCP stream reassembly */
static void tcpstream_decode_stream(void)
{
	struct tcp_session *s=cur.s;
	size_t t, dlen;
	void *dptr;

	/* TODO: Can only deliver data to user if rcv is in one of
	 * ESTABLISHED, FIN_WAIT1 or FIN_WAIT2. That means we have
	 * to hold on to any data until connection establishment.
	 */

	/* If the packet has no valid data then just shove it straight
	 * up the stack.
	 */
	if ( !(cur.ptr && cur.tp.len) )
		goto nodecode;

	/* See if we have already started queing this stream */
	if ( cur.rcv->queue ) {
		tcp_stream_queue();
		goto nodecode;
	}

	/* See if we need to start queing or not */
	t = s->proto->sdecode(cur.pkt, cur.ptr, cur.tp.len);

	/* If we are destroying the stream anyway, don't
	 * bother with the possible allocation of tcp
	 * reassembly buffer
	 */
	if ( t == 0 && cur.free )
		goto nodecode;

	/* If there was some valid data then we must proceed up
	 * to the decoder
	 */
	if ( t )
		tcpstream_decode();

	/* If the decoder took all the data in the packet, then there
	 * is no need to do any reassembly. This is the fast path.
	 */
	if ( likely(t >= cur.tp.len) )
		return;

	/* OK, some of the packet was un-processed, that
	 * means we need to start reassembling
	 */

	/* Skip over bit that was taken, if any */
	dlen = cur.tp.len - t;
	dptr = cur.ptr + t;

	/* Allocate the buffer */
	tcp_buffer_resize(cur.rcvbuf, dlen);
	if ( cur.rcvbuf->buf_sz < dlen ) {
		if ( t )
			return;
		goto nodecode;
	}

	/* All done */
	cur.rcv->queue = 1;
	memcpy(cur.rcvbuf->buf, dptr, dlen);
	dmesg(M_DEBUG, "tcpstream: start queuing %u/%u",
		cur.rcvbuf->buf_ofs, cur.rcvbuf->buf_sz);

nodecode:
	tcpstream_nodecode(cur.pkt);
	return;
}

/* Do some basic sanity checks to decide whether to even
 * bother processing a given segment or not.
*/
void tcpstream_process(struct packet *pkt, int i)
{
	struct pkt_iphdr *iph=pkt->layer[i].h.ip;
	struct layer *l=&pkt->layer[i+1];

	/* Indicate that tcp state tracking is on */
	l->flags|=FLAG_TCP_TRACK;

	cur.free = 0;

	/* Ignore fragments */
	if ( iph->frag_off & ipfmask )
		goto nodecode;

	tcp_packets++;

	/* Ignore bad IP checksums - ooh naughty! */
	if ( !(pkt->layer[i].flags & FLAG_IP_CSUM) )
		goto nodecode;

	/* Ignore bad TCP checksums - bad boy! */
	if ( !(l->flags & FLAG_TCP_CSUM) )
		goto nodecode;

	/* Ignore broadcast and multicast packets */
	if ( (pkt->flags & (FP_MULTICAST|FP_BROADCAST)) ) {
		tcp_broadcast++;
		goto nodecode;
	}

	/* Ignore packets with TTLs lower than minttl */
	if ( iph->ttl < tcp_minttl ) {
		tcp_lowttl++;
		goto nodecode;
	}

	/* Check we aren't in a icmp error message */
	if ( i>0 && pkt->layer[i-1].proto == &icmp_p )
		goto nodecode;

	/* Host order header values */
	cur.iph = pkt->layer[i].h.ip;
	cur.tcph = pkt->layer[i+1].h.tcp;
	cur.tp.ack = ntohl(cur.tcph->ack);
	cur.tp.seq = ntohl(cur.tcph->seq);
	cur.tp.win = ntohs(cur.tcph->win);
	cur.tp.len = ntohs(cur.iph->tot_len)
		- (cur.iph->ihl<<2)
		- (cur.tcph->doff<<2);
	cur.tp.seq_end = cur.tp.seq + cur.tp.len
		+ cur.tcph->flags.bits.syn
		+ cur.tcph->flags.bits.fin;
	cur.tp.tsval = 0;
	cur.tp.saw_tstamp = 0;
	cur.s = NULL;
	cur.ptr = NULL;
	cur.pkt = pkt;
	cur.layer = i;

	/* Actually do the state tracking */
	tcpstream_tcpseg(i);

	/* Keep track of statistics */
	if ( !(l->flags & FLAG_TCP_STATE) )
		tcp_state_errs++;

	/* Have we actually been decoded? */
	if ( !cur.s || !cur.s->proto )
		goto nodecode;

	if ( tcp_reassemble ) {
		tcpstream_decode_stream();
	}else{
		tcpstream_decode();
	}

done:
	if (cur.free)
		tcp_free(l->session);
	tcp_tmo_check(pkt);
	return;

nodecode:
	tcpstream_nodecode(pkt);
	goto done;
}

/* Called before firestorm shuts down */
void tcpstream_free(void)
{
	mesg(M_INFO,"tcpstream: max_concurrent=%u num_active=%u",
		max_concurrent, num_active);

	if ( tcp_reassemble ) {
		mesg(M_INFO,"tcpstream: max_flows=%u num_flows=%u",
			max_flows, num_flows);
	}

	mesg(M_INFO,"tcpstream: %u state errors out of %u packets",
	       tcp_state_errs, tcp_packets);
	mesg(M_INFO,"tcpstream: %u broadcasts, %u ttl evasions, "
		"%u timeouts, %u reassembled",
		tcp_broadcast, tcp_lowttl,
		tcp_timeouts, tcp_reassembled);

	if ( tcp_hash )
		free(tcp_hash);

	if ( all_sessions )
		free(all_sessions);
}

/* Initialise flow cache */
static int tcpstream_flow_init(void)
{
	void *p,*n=NULL;
	unsigned int i;

	/* Pointless if not reassembling, or have
	 * no application layer protocols with flow
	 * capabilities... */
	if ( !tcp_reassemble || !flow_len )
		return 0;

	/* not likely */
	if ( flow_len < sizeof(void *) ) {
		flow_len=sizeof(void *);
	}

	/* flows are 1 to 1 with streams, even in worst case */
	if ( tcp_numflows > tcp_numstreams ) {
		tcp_numflows=tcp_numstreams;
		mesg(M_WARN, "tcpstream: num_flows>num_streams, duh");
	}

	if ( tcp_numflows < 256 ) {
		mesg(M_WARN, "tcpstream: you may be at risk with "
			"such a low num_flows value");
	}

	/* Allocate it */
	if ( !(flow_cache=calloc(tcp_numflows, flow_len)) ) {
		return -1;
	}

	/* Initialise it */
	for(i=0; i<(tcp_numflows-1); ) {
		p=(char *)flow_cache + (i*flow_len);
		n=(char *)flow_cache + (++i*flow_len);
		*(void **)p = n;
	}

	*(void **)n=NULL;
	flow_next=flow_cache;

	return 0;
}

/* Initialise based on user parameters */
int tcpstream_init(const char *str, preproc_fn fn, void *priv)
{
	unsigned int i;

	if ( tcp_stateful ) {
		mesg(M_ERR,"tcpstream: Can't add tcpstream twice!");
		return 0;
	}

	if ( str ) {
		switch(args_parse(args, str, NULL)) {
		case -1:
			mesg(M_ERR,"tcpstream: parse error: %s", str);
		case 0:
			return 0;
		default:
			break;
		}
	}

	if ( tcp_minttl > 255 ) {
		mesg(M_ERR,"tcpstream: minttl must be < 256");
		return 0;
	}

	if ( tcp_numstreams < 32 ) {
		mesg(M_WARN,"tcpstream: num_streams is stupidly "
			"low, setting to 512");
		tcp_numstreams=512;
	}else if ( tcp_numstreams < 512 ) {
		mesg(M_WARN,"tcpstream: num_streams is VERY low! "
			"are you sure?!");
	}

	/* Select a size for the hash table such that
	 * num_streams/hash_size is no larger than 128
	 * and such that the minimum size is 512 */
	for(tcp_hashsz=512;
		tcp_numstreams/tcp_hashsz>128;
		tcp_hashsz<<=1);

	/* Make sure its always odd */
	tcp_hashsz|=1;

	/* Initialise data structures */
	if ( !(all_sessions=calloc(tcp_numstreams,
		sizeof(struct tcp_session))) ) {
		goto oom;
	}
	if ( !(tcp_hash=calloc(tcp_hashsz,
		sizeof(*tcp_hash))) ) {
		free(all_sessions);
		all_sessions=NULL;
		goto oom;
	}

	/* Initialise the slab-style allocator */
	for(i=0; i<(tcp_numstreams-1); i++) {
		all_sessions[i].next=&all_sessions[i+1];
	}
	all_sessions[i].next=NULL;
	tcp_next=all_sessions;

	/* Initialise state info */
	if ( tcp_reassemble ) {
		size_t i, flen;

		for(i=0; i<tcp_p.num_child; i++)
		{
			flen = tcp_p.child[i].proto->flow_len;
			if ( flen > flow_len )
				flow_len = flen;
		}
	}

	/* Initialise flow cache */
	if ( tcpstream_flow_init() )
		goto oom;

	/* Print out some statistics */
	mesg(M_INFO,"tcpstream: %u streams in %u buckets (%u KB)",
		tcp_numstreams, tcp_hashsz,
		(tcp_numstreams*sizeof(struct tcp_session))/1024);
	mesg(M_INFO,"tcpstream: TCP stream reassembly is %s: %u flows",
		tcp_reassemble ? "ENABLED" : "DISABLED",
		tcp_reassemble ? tcp_numflows : 0);

	tcp_stateful=1;

	return 1;
oom:
	mesg(M_ERR,"tcpstream: calloc(): %s", sys_err());
	return 0;
}
