/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/
#include <firestorm.h>
#include <dotscara/fdctl.h>
#include <dotscara/fobuf.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdarg.h>

#include <strtouint.h>
#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <exporter.h>
#include <plugin.h>

/* Private data structure */
struct ascii_priv {
	struct fobuf buf;
	int hex;
	unsigned int llen;
};

/* Arguments */
static struct arg ascii_args[]={
	{"file", ARGTYPE_STRING, .help = "Output filename"},
	{"hex", ARGTYPE_BOOL, .def = "true", .help = "Enable hex dump"},
	{"len", ARGTYPE_BYTES, .def = "16", .help = "Hex dump line length"},
	{NULL,}
};

static int print_error;
static char *abuf;
static size_t abuflen;

static void ascii_printf(struct fobuf *f, const char *fmt, ...)
	__printf(2,3) __nonull(1,2);
static void ascii_printf(struct fobuf *f, const char *fmt, ...)
{
	int len;
	va_list va;
	char *new;

again:
	va_start(va, fmt);

	len = vsnprintf(abuf, abuflen, fmt, va);
	if ( len < 0 ) /* bug in old glibc */
		len = 0;
	if ( (size_t)len < abuflen )
		goto done;

	new = realloc(abuf, len + 1);
	if ( new == NULL )
		goto done;

	abuf = new;
	abuflen = len + 1;
	goto again;

done:
	if ( !fobuf_write(f, abuf, (size_t)len) )
		print_error = 1;

	va_end(va);
}

static void hex_dump(struct fobuf *f, char *ptr,
			unsigned int len, unsigned int llen)
{
	unsigned int i, j;
	unsigned int line;
	unsigned char *tmp=(unsigned char *)ptr;

	for(j=0; j<len; j+=line, tmp+=line) {
		if ( j+llen > len ) {
			line=len-j;
		}else{
			line=llen;
		}

		ascii_printf(f, "%05X : ", j);

		for(i=0; i<line; i++) {
			if ( isprint(tmp[i]) ) {
				ascii_printf(f, "%c", tmp[i]);
			}else{
				ascii_printf(f, ".");
			}
		}

		for(; i<llen; i++)
			ascii_printf(f, " ");

		for(i=0; i<line; i++)
			ascii_printf(f, " %02X", tmp[i]);

		ascii_printf(f, "\n");
	}
}

static int ascii_queue(struct sink *s, struct packet *pkt)
{
	static char ln[256];
	struct ascii_priv *p = s->priv;
	struct alert *a = pkt->alert;
	struct tm rtime;
	int i;

	if ( !pkt )
		return 0;

	print_error = 0;

	time_to_local(pkt->time, &rtime);
	strftime(ln, sizeof(ln), "%Y-%m-%d %H:%M:%S", &rtime);

	/* Basic packet information */
	ascii_printf(&p->buf, "  packet: %s.%.9u len=%u caplen=%u\n",
		ln, time_nsec(pkt->time),
		pkt->len, pkt->caplen);

	/* Alert information */
	if ( a )
		ascii_printf(&p->buf, "   alert: [%s] %s (sid=%u.%u prio=%u)\n",
			a->gen->name, a->alert,
			a->sid, a->rev, a->priority);

	/* Print all the layers */
	for(i=0; i<pkt->llen; i++) {
		if ( !pkt->layer[i].proto ) {
			ascii_printf(&p->buf, "    data: Application "
				"layer data (%u bytes)\n",
				pkt->end-pkt->layer[i].h.raw);
			continue;
		}

		if ( pkt->layer[i].proto->print ) {
			pkt->layer[i].proto->print(pkt, i, ln, sizeof(ln));
		}else{
			sprintf(ln, "(no information)");
		}

		ascii_printf(&p->buf, "%8s: %s\n",
			pkt->layer[i].proto->name,
			ln);
	}

	/* hex dump the mofo */
	if ( p->hex ) {
		hex_dump(&p->buf, pkt->base, pkt->caplen, p->llen);
	}

	ascii_printf(&p->buf, "\n");

	if ( print_error || !fobuf_flush(&p->buf) ) {
		mesg(M_ERR, "ascii: output error: %s", sys_err2(NULL));
		return 0;
	}

	return 1;
}

static int ascii_init(struct sink *s, const char *args)
{
	struct ascii_priv *p;
	char *fn;

	if ( !(p=calloc(1, sizeof(*p))) )
		return 0;

	switch ( args_parse(ascii_args, args, p) ) {
	case -1:
		mesg(M_ERR,"ascii: parse error: %s", args);
	case 0: /* fall through */
		goto err;
	default:
		break;
	}

	fn = arg_get_string(&ascii_args[0].val.v_str);
	if ( ascii_args[0].val.v_str.len ) {
		int fd;

		fd = open(fn, O_WRONLY|O_CREAT|O_TRUNC, 0600);
		if ( fd < 0 ) {
			mesg(M_ERR, "ascii: %s: open(): %s", fn, sys_err());
			goto err;
		}

		if ( !fobuf_attach(&p->buf, fd, 0, 0) ) {
			fd_close(fd);
			goto err;
		}
	}else{
		if ( !fobuf_attach(&p->buf, STDOUT_FILENO, 0, 0) )
			goto err;
	}

	p->hex = ascii_args[1].val.v_bool;
	p->llen = ascii_args[2].val.v_uint;

	s->priv = p;
	return 1;

err:
	free(p);
	return 0;
}

static int ascii_free(struct sink *s)
{
	struct ascii_priv *p = s->priv;
	int ret = 1;

	if ( !fobuf_close(&p->buf) ) {
		mesg(M_ERR, "fobuf_close(): %s", sys_err2(NULL));
		ret = 0;
	}

	free(p);
	return ret;
}

struct exporter firestorm_exporters[]={
	{
		.name = "ascii",
		.desc = "Detailed ascii text with optional hex dumps",
		.queue = ascii_queue,
		.init = ascii_init,
		.free = ascii_free,
		.args = ascii_args,
	},
	{.name = NULL},
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "exporter.ascii",
	.desc		= "Detailed ASCII logs",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 1,
	.ver_minor	= 1,
};
