/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/
#ifndef __FIRESTORM_HEADER_INCLUDED__
#define __FIRESTORM_HEADER_INCLUDED__

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#if HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include <time.h>

#include <dotscara/dotscara.h>
#include <dotscara/sys.h>
#include <dotscara/fobuf.h>
#include <dotscara/fibuf.h>

#include <list.h>

struct fvec {
	void *ptr;
	size_t len;
};

typedef u_int64_t timestamp_t;

typedef u_int8_t mesg_code_t;
#define M_UNSET	0 /* Unclassified */
#define M_CRIT	1 /* Service/data is lost */
#define M_ERR	2 /* We can't do something you asked for */
#define M_WARN	3 /* We can work around this, but you should know */
#define M_INFO	4 /* Informational notices */
#define M_DEBUG	5 /* For developers */
#define M_MAX	6
#define M_LIMIT 0x80 /* Ratelimit this message */

#include <idx.h>
#include <packet.h>
#include <notifier.h>
#include <nbio.h>
#include <pipeline.h>

/* Subsystem intialization */
void __private loader_init(void);
void __private notifier_init(void);
void __private decode_init(void);
void __private field_init(void);
void __private capture_init(void);

/* Core firestorm API */
void mesg(mesg_code_t code, const char *fmt, ...) __printf(2,3);
void notify_exit(notifier_t fn, void *priv) __nonull(1);
void firestorm_init(void);
void firestorm_exit(int, const char *, ...) __printf(2,3) __noreturn;

#define firestorm_die_error(x) \
	firestorm_exit(EXIT_FAILURE, "%s::%s:%i: " x "(): %s", \
		__FILE__, __FUNCTION__, __LINE__, sys_err2(NULL));

/* Firestorm packet handling API */
void packet_reinject_prepare(struct packet *dst, struct packet *src)
	__nonull(1,2);

/* Firestorm time API - Firestorm timestamps are 64bit unsigned nanosecond
 * resolution counters, starting from the UNIX epoch of 0000 Jan 1 1970. They
 * won't wrap until about 2555.
*/
#define TIMESTAMP_INFINITE 0xffffffffffffffffULL
#define TIMESTAMP_HZ  1000000000ULL
#define TIMESTAMP_KHZ 1000000ULL
#define TIMESTAMP_MHZ 1000ULL

/* get the current system time (usually you do not want to use this) */
timestamp_t time_gettime(void);

/* get virtual timestamp (cpu time) */
timestamp_t time_getvtime(void);

/* do not use this directly */
extern timestamp_t firestorm_now;

/* get the current firestorm time */
static inline timestamp_t time_now(void)
{
	return firestorm_now;
}

static inline u_int64_t time_secs(timestamp_t t)
{
	return (t / TIMESTAMP_HZ);
}

static inline u_int32_t time_nsec(timestamp_t t)
{
	return (u_int32_t)(t % TIMESTAMP_HZ);
}
/* set the current firestorm time */
static inline void time_update_now(timestamp_t new_now)
{
	firestorm_now = new_now;
}

/* return 1 if t1 is before t2 (wrap safe) */
static inline int time_before(timestamp_t t1, timestamp_t t2)
{
	return (int64_t)(t1-t2) < 0;
}

/* return 1 if t1 is after t2 (wrap safe) */
static inline int time_after(timestamp_t t1, timestamp_t t2)
{
	return (int64_t)(t2-t1) < 0;
}

/* conversions */
time_t time_to_time_t(timestamp_t);
void time_to_local(timestamp_t, struct tm *) __nonull(2);
void time_to_gmt(timestamp_t, struct tm *) __nonull(2);
void time_to_timeval(timestamp_t, struct timeval *) __nonull(2);
timestamp_t time_gcd(timestamp_t n, timestamp_t d);

static inline timestamp_t time_from_timeval(struct timeval *) __nonull(1);
static inline timestamp_t time_from_timeval(struct timeval *tv)
{
	return (tv->tv_sec * TIMESTAMP_HZ) +
		(tv->tv_usec * TIMESTAMP_MHZ);
}

/* Generic vector API */
int vcasecmp(const struct fvec *v1, const struct fvec *v2) __nonull(1,2);
int vcmp(const struct fvec *v1, const struct fvec *v2) __nonull(1,2);
int vstrcmp(const struct fvec *v1, const char *str) __nonull(1,2);
size_t vtouint(struct fvec *v, unsigned int *u) __nonull(1,2);

/* Eeasy explode() type function for parsing */
int easy_explode(char *str, char split, char **toks, int max_toks);

#define RATE_SEC  (TIMESTAMP_HZ)
#define RATE_MIN  (60ULL*RATE_SEC)
#define RATE_HR   (60ULL*RATE_MIN)
#define RATE_DAY  (24ULL*RATE_HR)
#define RATE_WEEK (7ULL*RATE_DAY)
#define RATE_YEAR (365ULL*RATE_DAY)

struct tokenbucket {
	timestamp_t last_msg; /* time of last msg */
	timestamp_t toks; /* tokens */
	timestamp_t cost; /* max rate */
	timestamp_t burst; /* burst */
	int missed; /* number of missed alerts */
};

/* Nanosecond resolution token bucket rate-limiting filter.
 * Return values:
 *  o -1 for a drop
 *  o 0 <= n < INT_MAX for a pass, n indicates number of drops
 *    since last pass
 */
int token_bucket(struct tokenbucket *a, timestamp_t h);

#endif /* __FIRESTORM_HEADER_INCLUDED__ */
