/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/
#ifndef __PACKET_HEADER_INCLUDED__
#define __PACKET_HEADER_INCLUDED__

/* packet macros */
#include <pmacros.h>

/* Various protocol headers */
#include <pkt/eth.h>
#include <pkt/ip.h>
#include <pkt/icmp.h>
#include <pkt/igmp.h>
#include <pkt/vlan.h>
#include <pkt/tcp.h>
#include <pkt/udp.h>
#include <pkt/arp.h>
#include <pkt/gre.h>
#include <pkt/sll.h>
#include <pkt/http.h>
#include <pkt/ipx.h>
#include <pkt/sap.h>

/* Maximum number of layers in decode */
#define PKT_LAYERS	16

struct layer {
	/* Various protocols are supported in the
	 * union by default - its not really a modular
	 * violation - it just makes things easier to follow */
	union {
		struct pkt_ethhdr	*eth; /* Ethernet II */
		struct pkt_machdr	*mac; /* 802.3 MAC */
		struct pkt_llchdr	*llc; /* Logical Link Control */
		struct pkt_snaphdr	*snap; /* Sub-Network Access Protocol */
		struct pkt_iphdr	*ip;
		struct pkt_icmphdr	*icmp;
		struct pkt_igmphdr	*igmp;
		struct pkt_tcphdr	*tcp;
		struct pkt_arphdr	*arp;
		struct pkt_udphdr	*udp;
		struct pkt_grehdr	*gre;
		struct pkt_vlanhdr	*vlan;
		struct pkt_sllhdr	*sll;
		struct pkt_ipxhdr	*ipx;
		struct pkt_sap		*sap;
		void			*raw;
	}h;
	struct proto	*proto; /* protocol of this layer */
	u_int32_t	flags; /* protocol specific flags */
	void		*session; /* decode or session information */
};

struct packet {
	/* length of packet on the wire */
	size_t		len;

	/* Length of captured packet */
	size_t		caplen;
	
	/* decode information */
	struct layer	layer[PKT_LAYERS];
	int		llen;

	/* Actual packet data */
	void		*base;
	void		*end;

	/* time of capture */
	timestamp_t	time;

#define FP_WRITABLE	(1<<0)  /* Packet is writable */
#define FP_COPY		(1<<1)  /* memcpy to clone the payload? */
#define FP_REAL		(1<<2)  /* Real packet: ie, not from refragmentation */
#define FP_HOST		(1<<3)  /* Directly to sensor */
#define FP_MULTICAST	(1<<4)  /* Multicast packet */
#define FP_BROADCAST	(1<<5)  /* Broadcast packet */
#define FP_PROMISC	(1<<6)  /* Promiscuous packet */
#define FP_OUTGOING	(1<<7)  /* Sensor generated */
#define FP_PKTTYPE      (1<<8)  /* packet type is known for sure */
#define FP_BIG_ENDIAN	(1<<9)  /* Endianness */
	u_int32_t	flags;

	/* Alert information, if any */
	struct alert	*alert;

	/* Packet source is always set */
	struct source	*source;
};

/* If unsure, about packet type just select PROMISC ;) */
#define FPMASK_PKTTYPE (FP_HOST|FP_MULTICAST|FP_BROADCAST|\
			FP_PROMISC|FP_OUTGOING)

/* Some flags may not be saved/loaded from files */
#define FPMASK_NON_PERSISTANT (FP_WRITABLE|FP_COPY)
#define FPMASK_PERSISTANT (~FPMASK_NON_PERSISTANT)

#if __BYTE_ORDER == __LITTLE_ENDIAN
#define HOST_ORDER 0
#elif __BYTE_ORDER == __BIG_ENDIAN
#define HOST_ORDER FP_BIG_ENDIAN
#else
#error "Couldn't determine endianness"
#endif

static inline u_int32_t pkt32(struct packet *p, u_int32_t val)
{
	if ( (p->flags & FP_BIG_ENDIAN) ^ HOST_ORDER )
		return sys_bswap32(val);
	return val;
}
static inline u_int16_t pkt16(struct packet *p, u_int16_t val)
{
	if ( (p->flags & FP_BIG_ENDIAN) ^ HOST_ORDER )
		return sys_bswap16(val);
	return val;
}

#endif /* __PACKET_HEADER_INCLUDED__ */
