/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2004 Gianni Tedesco
 * Released under the terms of the GNU GPL version 2
*/
#ifndef __PIPELINE_HEADER_INCLUDED__
#define __PIPELINE_HEADER_INCLUDED__

#ifdef __GNUC__
struct q_filter;
#endif

#define HOOK_PRE_PROCESSING	0 /* modify packets before matching */
#define HOOK_PROCESSING		1 /* do some detection, tag alerts */
#define HOOK_POST_PROCESSING	2 /* get all packets, alerts or no */
#define HOOK_PRE_ALERT		3 /* alert packets, before token bucket */
#define HOOK_ALERT		4 /* alert packets */
#define HOOK_MAX		5

#define SINK_ALL		0 /* All packets */
#define SINK_REAL		1 /* Only real packets */
#define SINK_PRE_ALERT		2 /* All alerted packets */
#define SINK_ALERT		3 /* All alerted packets (rate limited) */
#define SINK_MAX		4

#define PIPELINE_STATUS_STOPPED	0
#define PIPELINE_STATUS_INIT	1
#define PIPELINE_STATUS_RUNNING	2

struct intr_ops {
	void (*lock)(void *);
	void (*unlock)(void *);
};

struct source {
	struct list_head list;
	struct packet *last_pkt;
	struct capdev *plugin;
	void *priv;
	struct pipeline *owner;
	char args[0];
};

struct sink {
	struct list_head list;
	struct exporter *plugin;
	void *priv;
	struct pipeline *owner;
	char args[0];
};

struct pipeline {
	struct iothread iothread;
	bitmask_t flags;
	u_int64_t tot_bytes;
	int status;
	struct list_head f_sources; /* file */
	struct list_head l_sources; /* live */
	struct list_head d_sources; /* deleted */
	struct list_head sinks;
	unsigned int num_f_sources;
	unsigned int num_l_sources;
	struct list_head hook[HOOK_MAX];
	struct list_head ohook[SINK_MAX];
	struct per_proto *proto;
};

/* Indicates live captures may be added. If any live captures are added
 * then nbio will be performed */
#define PIPELINE_ALLOW_LIVE		(1<<0)

/* Indicates file captures may be added */
#define PIPELINE_ALLOW_FILE		(1<<1)

/* Indicates that any files which are added must be merged rather
 * than processed one-at-a-time. This flag is meaningless if
 * PIPELINE_ALLOW_FILE is not set.
 */
#define PIPELINE_MERGE			(1<<2)

/* Quit the mainloop if any of the sinks returns an error */
#define PIPELINE_SINK_ERRORS_FATAL	(1<<3)

/* Pause for signals before capture starts and when capture ends */
#define PIPELINE_PROFILE_MODE		(1<<4)

/* Possible interrupts - interrupts effect all pipelines */
#define PIPELINE_INTR_SPURIOUS		(1<<0)
#define PIPELINE_INTR_USR1		(1<<1)
#define PIPELINE_INTR_USR2		(1<<2)
#define PIPELINE_INTR_HUP		(1<<3)
#define PIPELINE_INTR_ITIMER		(1<<4)
#define PIPELINE_INTR_VTIMER		(1<<4)
#define PIPELINE_INTR_DEADLY		(1<<5)

/* Set interrupt operations */
void pipeline_intr_ops(struct intr_ops *ops, void *priv);

/* Send asynchronous interrupts to all pipelines */
void pipeline_intr(bitmask_t intrs, int reentran);

/* Initialises a pipeline */
int pipeline_init(struct pipeline *p, bitmask_t flags) __nonull(1);

/* Frees everything associated with a pipeline, outputs files will
 * all be closed and error state ignored.
 */
void pipeline_free(struct pipeline *p) __nonull(1);

/* Add a file or live-capture source */
struct source *pipeline_source_add(struct pipeline *p,
					const char *plugin,
					const char *args) __nonull(1,2);

/* Add a file source using format-recognition */
struct source *pipeline_source_add_file(struct pipeline *p,
					struct fibuf *fi,
					const char *fn)
					__nonull(1,2,3);

/* Returns 1 if nbio will happen */
int pipeline_is_live(struct pipeline *p) __nonull(1);

/* Add a preprocessor */
int pipeline_preproc_add(struct pipeline *p, const char *plugin,
				const char *args) __nonull(1,2);

/* Add a sink */
struct sink *pipeline_sink_add(struct pipeline *p,
				int h, /* HOOK_XXX */
				const char *plugin,
				const char *args) __nonull(1,3);

/* Close a sink, or all sinks */
int pipeline_sink_close(struct sink *s) __nonull(1);
int pipeline_sink_close_all(struct pipeline *p) __nonull(1);

/* process a pipeline */
int pipeline_go(struct pipeline *p) __nonull(1);

/* Rewind all file sources */
void pipeline_rewind(struct pipeline *p) __nonull(1);

/* Add a notifier callback to a given hook */
struct notifier *pipeline_hook_add(struct pipeline *p,
				int h, /* SINK_XXX */
				notifier_t n,
				notifier_dtor_t d,
				void *priv) __nonull(1);

/* Delete a given notifier */
void pipeline_hook_cancel(struct notifier *n) __nonull(1);

/* Delete all notifiers on a given hook */
void pipeline_hook_delete(struct pipeline *p, int h) __nonull(1);

/* Delete all notifiers */
void pipeline_hook_delete_all(struct pipeline *p) __nonull(1);

/* Called by decoders to dispatch packet */
void dispatch(struct packet *pkt) __nonull(1);

#endif /* __PIPELINE_HEADER_INCLUDED__ */
