/******************************************************************************/
/* B.A.D. v0.3 - Bottleneck Anomoly Detector                                  */
/* (c) 2003 jamie twycross, milieu3@ukshells.co.uk                            */
/* released under gnu gpl version 2                                           */
/******************************************************************************/

#include <firestorm.h>

#include <limits.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include "network.h"
#include "dataset.h"

#include <plugin.h>
#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <preproc.h>
#include <alert.h>

#ifdef DEBUG
#define dbg mesg
#else
#define dbg(...)
#endif

#ifdef BENCHMARK
#define bmark mesg
#else
#define bmark(...)
#endif

#ifdef STATS
#define stats mesg
#else
#define stats(...)
#endif

/***********/
/* globals */
/***********/

static unsigned int hidden_layer_size;  /* num of nodes in bottleneck */ 
static unsigned int training_set_size;  /* num of vectors in training set */ 
static unsigned int training_iterations;  /* steps to train network for */
static double refresh_rate;  /* rate at which to retrain network */
static double learning_rate;  /* network learning rate */
static double momentum;  /* network momentum coefficient */

static network_t *network;  /* our bottleneck network */ 
static dataset_t *training_set;  /* set to train network on */
static short unsigned int training_flag;  /* initially training network */
static double error_threshold;  /* error threshold of current network */
static unsigned int vectors_needed;  /* current num of training examples */
static unsigned int refresh_vectors;  /* num of vectors before retraining */

/**********/
/* alerts */
/**********/

static struct tokenbucket bad_tb = {
	.cost = (RATE_SEC / 1),  /* 1 alert per second */
	.burst = (RATE_SEC / 1) * 10  /* burst of 10 alerts per second */
};

struct generator bad_gen = init_generator("bad", &bad_tb);

static struct alert alert_bad1 = {
	.alert = "Anomolous TCP packet",
	.sid = 1, .rev = 0,
	.priority = 5,
	.gen = &bad_gen,
	.order = ALERT_BUILTIN,
};

/******************************/
/* packet processing routines */
/******************************/

/* filter packet stream for interesting packets */
static unsigned int filter_packet(struct packet *pkt, int l)
{
	/* filter for packets indicating establishment of tcp connection */
	if(pkt->layer[l].flags & FLAG_TCP_CT_EST)
		return 1;
	return 0;
}

#define FV_SIZE 10  /* dimension of feature vector */

#define byte(a,b) (*(((unsigned char *) &(a)) + (b)))

/* convert packet to feature vector */
static void create_feature_vector(struct packet *pkt, int l, double *fv)
{
	register struct tcp_session *s = pkt->layer[l].session;
	
	/* currently looking at src/dst ip addr and tcp port */
	fv[0] = (double) byte(s->s_addr, 0);  /* server addr */
	fv[1] = (double) byte(s->s_addr, 1);
	fv[2] = (double) byte(s->s_addr, 2);
	fv[3] = (double) byte(s->s_addr, 3);
	fv[4] = (double) byte(s->s_port, 0);  /* server port */
	fv[5] = (double) byte(s->s_port, 1);
	fv[6] = (double) byte(s->c_addr, 0);  /* client addr */
	fv[7] = (double) byte(s->c_addr, 1); 
	fv[8] = (double) byte(s->c_addr, 2);
	fv[9] = (double) byte(s->c_addr, 3);
//	fv[10] = (double) byte(s->c_port, 0);  /* client port */
//	fv[11] = (double) byte(s->c_port, 1);
}

/* train network for given num of iterations */
static void train_network(network_t *net, dataset_t *ds, const unsigned int its)
{
	register unsigned int i, j;
	
	bmark(M_INFO, "bad: started training network");
	/* normalise dataset */
	dataset_normalise(ds); 
	/* train for specified num of iterations */
	for(i = 0; i < its; i++) {
		/* change presentation order */
		dataset_permute(ds);  		
		/* go thru training set once */
		for(j = 0; j < ds->nvecs; j++) {
			/* apply training vector to network */
			net_compute(net, ds->normvec[ds->order[j]], NULL); 
			/* calculate error */
			net_compute_output_error(net,
					ds->normvec[ds->order[j]]); 
			/* train one step */
			net_train(net);
		}
	}
	bmark(M_INFO, "bad: finished training network");
}

/* evaluate feature vector and return mean square error */
static inline double evaluate_feature_vector(network_t *network,
		dataset_t *ds, double *fv)
{
	static double inputs[FV_SIZE];
	register double error;
	
	bmark(M_INFO, "bad: started evaluating vector");
	/* normalise fv */
	dataset_normalise_vector(ds, fv, inputs);
	/* apply feature vector to network */
	net_compute(network, inputs, NULL);
	/* calculate error */
	error = 2.0 * net_compute_output_error(network, inputs) / ds->ncomps;
	bmark(M_INFO, "bad: finished evaluating vector");
	stats(M_INFO, "bad: error=%f", error);

	return error;
}

/* calculate error threshold of trained network */
/* currently calculates threshold as mean of mean square errors of the */
/* network on the training set plus one standard deviation */
/* smarter ways to do this */
static double calculate_error_threshold(network_t *net, dataset_t *ds)
{
	register unsigned int i;
	register double mean, sd;
	
	bmark(M_INFO, "bad: started calculating error threshold");
	/* calculate mean of global error on training set */
	mean = 0.0;
	for(i = 0; i < ds->nvecs; i++) {
		/* apply training vector to network */
		net_compute(net, ds->normvec[i], NULL); 
		/* calculate error and save */
		ds->dtmp[i] = 2.0 * net_compute_output_error(net,
				ds->normvec[i]) / ds->ncomps;
		mean += (ds->dtmp[i] / ds->nvecs); 
	}

	/* calculate standard deviation of error on training set */
	sd = 0.0;
	for(i = 0; i < ds->nvecs; i++)
		sd += (((ds->dtmp[i] - mean) * (ds->dtmp[i] - mean)) /
				ds->nvecs);
	sd = sqrt(sd);
	bmark(M_INFO, "bad: finished calculating error threshold");
	dbg(M_INFO, "bad: error threshold: mean=%f sd=%f threshold=%f",
			mean, sd, mean + sd);
	stats(M_INFO, "bad: mean=%f sd=%f", mean, sd);
	
	/* return mean plus one standard deviation */
	return (mean + sd);  
}

/* process a packet */
/* try to return as quickly as possible */
static void bad_process(struct packet *pkt, int l, void *priv)
{
	static double fv[FV_SIZE];
	register double error;

	/* decide if packet interesting */
	if(!filter_packet(pkt, l))
		return;
	
	/* make a feature vector from the packet */
	create_feature_vector(pkt, l, fv);
	
	/* are we initially training the network? */
	if(training_flag) {
		/* yes, so add to training set */
		dataset_addvec(training_set, fv, --vectors_needed);
		/* got complete training set? */
		if(vectors_needed)
			return;  /* no, so done */
		/* yes, so train network */
		train_network(network, training_set, training_iterations);
		/* and calculate new error threshold */
		error_threshold = calculate_error_threshold(network,
				training_set);
		/* and setup for retraining */
		training_flag = 0;
		vectors_needed = refresh_vectors;
		return;
	}

	/* evaluate packet */
	error = evaluate_feature_vector(network, training_set, fv);
	
	/* anomolous? */
	if(error > error_threshold) {
		/* yes, so generate alert */
		dbg(M_INFO, "bad: packet anomolous: error=%f", error);
		alert_tag(pkt, &alert_bad1, -1);
		return;
	}

	dbg(M_INFO, "bad: packet not anomolous: error=%f", error);
	/* add to training set? */
	if(!vectors_needed)
		return;  /* no, so done */
	/* add to training set at random index */
	dataset_addvec(training_set, fv, (unsigned int) (training_set_size *
				rand() / (RAND_MAX + 1.0)));

	// got complete training set?
	if(--vectors_needed)
		return;  /* no, so done */
	/* yes, so retrain network */
	train_network(network, training_set, training_iterations);
	/* and calculate new error threshold */
	error_threshold = calculate_error_threshold(network, training_set);
	/* setup for next retraining */
	vectors_needed = refresh_vectors;
	
	return;
}

/****************************/
/* initialisiation routines */
/****************************/

#define is_prob(a)	(((a) >= 0.0) && ((a) <= 1.0))

/* parse those preprocessor arguments... */
static int parse_args(const char *args)
{
	struct arg bad_args[] =	{
		{"hls", ARGTYPE_PUINT, {vp_uint:&hidden_layer_size}},
		{"tss", ARGTYPE_PUINT, {vp_uint:&training_set_size}},
		{"ti", ARGTYPE_PUINT, {vp_uint:&training_iterations}},
		{"rr", ARGTYPE_PDOUBLE, {vp_double:&refresh_rate}}, 
		{"lr", ARGTYPE_PDOUBLE, {vp_double:&learning_rate}},
		{"mom", ARGTYPE_PDOUBLE, {vp_double:&momentum}},
		{NULL},
	};
	
	/* set defaults */
	hidden_layer_size = 8;
	training_set_size = 100;
	training_iterations = 1000;
	refresh_rate = 0.5;
	learning_rate = 0.25;
	momentum = 0.08;
	
	/* let firestorm have a go */
	switch(args_parse(bad_args, args, NULL)) {
		case -1:
			mesg(M_ERR, "bad: parse error: %s", args);
		case 0: /* fall through */
			return 0;
		default:
			break;
	}

	/* sanity check args */
	if((hidden_layer_size <= 0) || (hidden_layer_size >= FV_SIZE)) {
		mesg(M_ERR, "bad: hidden layer size %u out of range",
				hidden_layer_size);
		return 1;
	}
	if(training_set_size <= 0) {
		mesg(M_ERR, "bad: training set size %u out of range",
				training_set_size);
		return 1;
	}
	if(training_iterations <= 0) {
		mesg(M_ERR, "bad: training iterations %u out of range",
				training_iterations);
		return 1;
	}
	if(!is_prob(refresh_rate)) {
		mesg(M_ERR, "bad: refresh rate %f out of range", refresh_rate);
		return 1;
	}
	if(!is_prob(learning_rate)) {
		mesg(M_ERR, "bad: learning rate %f out of range",
				learning_rate);
		return 1;
	}
	if(!is_prob(momentum)) {
		mesg(M_ERR, "bad: momentum %f out of range", momentum);
		return 1;
	}
	
	return 0;
}

/* shutdown anomoly detector */
static void bad_free(void *priv)
{
	net_free(network);
	dataset_free(training_set);
}

/* initialise anomoly detector */
static int bad_init(const char *args, preproc_fn fn, void *priv)
{
	struct list_head *n;

	n = (*fn)("tcp", priv);
	if ( n == NULL )
		return 0;

	/* should use more random rngs... */
	srand(time(NULL)); 
	srand48(time(NULL));
	
	/* parse preproc args */
	if(parse_args(args)) {
		mesg(M_ERR, "bad: couldn't parse arguments");
		return 0;
	}
	
	/* create network */
	if(!(network = net_allocate(3, FV_SIZE, hidden_layer_size, FV_SIZE))) {
		mesg(M_ERR, "bad: couldn't allocate network");
		return 0;
	}
	
	/* create training set */
	if(!(training_set = dataset_init(training_set_size, FV_SIZE))) {
		mesg(M_ERR, "bad: couldn't allocate training set");
		return 0;
	}

	/* set network parameters */
	net_set_learning_rate(network, learning_rate);
	net_set_momentum(network, momentum);
	
	/* setup training */
	training_flag = 1;
	vectors_needed = training_set_size;
	refresh_vectors = (unsigned int) ((double) refresh_rate *
			training_set_size);
	
	/* tell the user what's happening */
	mesg(M_INFO, "bad: setting hidden layer size to %u", hidden_layer_size);
	mesg(M_INFO, "bad: setting training set size to %u", training_set_size);
	mesg(M_INFO, "bad: setting training iterations to %u",
			training_iterations);
	mesg(M_INFO, "bad: setting refresh rate to %f", refresh_rate);
	mesg(M_INFO, "bad: setting learning rate to %f",
			net_get_learning_rate(network));
	mesg(M_INFO, "bad: setting momentum to %f", net_get_momentum(network));
	
	if ( !notifier_add(n, bad_process, bad_free, NULL) )
		return 0;

	return 1;
}

/*****************/
/* the api stuff */
/*****************/

struct preproc firestorm_preprocs[] = {
	{.name = "bad", .init = bad_init},
	{.name = NULL}
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "preproc.bad", 
	.desc		= "B.A.D. - Bottleneck Anomoly Detector",
	.author_name	= "Jamie Twycross", 
	.author_email	= "milieu3@ukshells.co.uk",
	.ver_major	= 0,
	.ver_minor	= 3
};
