/******************************************************************************/
/* generic dataset routines                                                   */
/* (c) 2003 jamie twycross, milieu3@ukshells.co.uk                            */
/* released under gnu gpl version 2                                           */
/******************************************************************************/

#include "dataset.h"

#include <stdlib.h>
#include <math.h>

/* allocate memory for dataset */
dataset_t *dataset_init(const unsigned int nvecs, const unsigned int ncomps)
{
	dataset_t *ds;
	unsigned int i;

	if(!(ds = malloc(sizeof(dataset_t))))
		return NULL;
	ds->nvecs = nvecs;
	ds->ncomps = ncomps;
	if(!(ds->vec = calloc(ds->nvecs, sizeof(double *))))
		return NULL;
	if(!(ds->normvec = calloc(ds->nvecs, sizeof(double *))))
		return NULL;
	if(!(ds->order = calloc(ds->nvecs, sizeof(unsigned int))))
		return NULL;
	if(!(ds->means = calloc(ds->ncomps, sizeof(double))))
		return NULL;
	if(!(ds->sds = calloc(ds->ncomps, sizeof(double))))
		return NULL;
	if(!(ds->dtmp = calloc(ds->nvecs, sizeof(double))))
		return NULL;
	for(i = 0; i < ds->nvecs; i++) {
		if(!(ds->vec[i] = calloc(ds->ncomps, sizeof(double))))
			return NULL;
		if(!(ds->normvec[i] = calloc(ds->ncomps, sizeof(double))))
			return NULL;
		ds->order[i] = i;
	}

	return ds;
}

/* deallocate memory for dataset */
void dataset_free(dataset_t *ds)
{
	unsigned int i;

	for(i = 0; i < ds->nvecs; i++) {
		free(ds->vec[i]);
		free(ds->normvec[i]);
	}
	free(ds->dtmp);
	free(ds->sds);
	free(ds->means);
	free(ds->order);
	free(ds->normvec);
	free(ds->vec);
	free(ds);
}

/* add vector to dataset */
inline void dataset_addvec(dataset_t *ds, double *vec, \
		const unsigned int index)
{
	register unsigned int i;

	for(i = 0; i < ds->ncomps; i++)
		ds->vec[index][i] = vec[i];
}

/* normalise vectors in dataset and compute new means and sds */
inline void dataset_normalise(dataset_t *ds)
{
	register unsigned int i, j;
	register double mean, sd;

	/* for each input variable */
	for(i = 0; i < ds->ncomps; i++) {
		/* find mean */
		mean = 0.0;
		for(j = 0; j < ds->nvecs; j++)
			mean += (ds->vec[j][i] / ds->nvecs);
		/* find standard deviation */
		sd = 0.0;
		for(j = 0; j < ds->nvecs; j++)
			sd += ((ds->vec[j][i] - mean) * (ds->vec[j][i] - mean) \
					/ ds->nvecs);
		sd = sqrt(sd);
		/* normalise each component */
		for(j = 0; j < ds->nvecs; j++)
			if(sd != 0.0)
				ds->normvec[j][i] = (ds->vec[j][i] - mean) / sd;
			else
				ds->normvec[j][i] = 0.0;
		/* save mean + sd */
		ds->means[i] = mean;
		ds->sds[i] = sd;
	}
}

/* normalise vector w.r.t. current dataset means and sds */
inline void dataset_normalise_vector(dataset_t *ds, double *ivec, double *ovec)
{
	register unsigned int i;

	for(i = 0; i < ds->ncomps; i++)
		if(ds->sds[i] != 0.0)
			ovec[i] = (ivec[i] - ds->means[i]) / ds->sds[i];
		else
			ovec[i] = ivec[i] - ds->means[i];
}
	
/* permute presentation order */
inline void dataset_permute(dataset_t *ds)
{
	register unsigned int i, index, tmp;

	for(i = 0; i < ds->nvecs; i++) {
		index = (unsigned int) ((double) ds->nvecs * rand() / \
				(RAND_MAX + 1.0));
		tmp = ds->order[index];
		ds->order[index] = ds->order[i];
		ds->order[i] = tmp;
	}
}

/* convert double vector to float vector */
inline float *dataset_vtof(float *fvec, double *dvec, const unsigned int len)
{
	register unsigned int i;

	for(i = 0; i < len; i++)
		fvec[i] = (float) dvec[i];
	return fvec;
}
