/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002,2003,2004 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This is the detection engine, it is generic to all protocols
* although each protocol is expected to maintain its own root
* node. It is a very fast decision tree. If you want more details
* mail me ;) -- GT
*
* TODO:
*  o Use a custom fixed-size on-heap stack for recursion
*  o Allow rule removal
*  o Use a centralized allocator for tree nodes with colouring
*    based on tree depth.
*  o Update plugins to use real IDD subnode searches where relevent
*/
#include <firestorm.h>
#include <alert.h>
#include <signature.h>
#include <idx.h>
#include <matcher.h>
#include <field.h>
#include <detect.h>

static struct detect *all_nodes;

int detect_optimize(void)
{
	struct detect *d;

	for(d=all_nodes; d; d=d->next) {
		if ( !d->type->optimize )
			continue;
		if ( !d->type->optimize(d) ) {
			mesg(M_ERR, "%s: failed to optimize",
				d->type->owner->name);
			return 0;
		}
	}

	return 1;
}

static struct detect *new_node(struct sig_node *sig)
{
	struct detect *new;

	if ( sig->match->new_node == NULL ) {
		mesg(M_CRIT, "%s: matcher needs converting "
				"to new API.", sig->match->owner->name);
		return NULL;
	}

	new = sig->match->new_node(sig->match);
	if ( new == NULL )
		return NULL;

	new->type = sig->match;
	new->next = all_nodes;
	all_nodes = new;

	return new;
}

static int cmp(const void *A, const void *B)
{
	struct sig_node *a = (struct sig_node *)A;
	struct sig_node *b = (struct sig_node *)B;

	if ( a->match->owner->cost < b->match->owner->cost )
		return -1;
	if ( a->match->owner->cost > b->match->owner->cost )
		return 1;
	if ( a->match < b->match )
		return -1;
	if ( a->match > b->match )
		return 1;
	return 0;
}

static int is_compatible(struct detect *n, struct sig_node *c)
{
	struct matcher *m1 = NULL, *m2 = NULL;

	if ( n->type == c->match )
		return 1;

	if ( n->type->owner->compatible )
		m1 = matcher_find(n->type->owner->compatible);
	if ( c->match->owner->compatible )
		m2 = matcher_find(c->match->owner->compatible);

	if ( m1 && m1 == c->match )
		return 1;

	if ( m2 && m2 == n->type )
		return 1;

	return 0;
}

int detect_add_sig(struct detect_root *r, struct sig_node *sig,
			unsigned int len, struct alert *alert)
{
	struct detect **cur_node = &r->child;
	struct sig_node *c;
	unsigned int i;

	qsort(sig, len, sizeof(*sig), cmp);

#if 0
	mesg(M_DEBUG, "Signature: %u criteria, (sid=%u rev=%u) - '%s'",
			len, alert->sid, alert->rev, alert->alert);
	for(i=0; i < len; i++) {
		mesg(M_DEBUG, " o (%s - %s)",
			sig[i].match->owner->name,
			sig[i].match->owner->desc);
	}
#endif

	sig[len-1].alert = alert;

	for(i=0,c=sig; i < len; i++,c++) {
do_something:
		if ( *cur_node == NULL ) {
			struct detect *new;

			new = new_node(c);
			if ( new == NULL )
				return 0;

			*cur_node = new;
			cur_node = new->type->insert_node(new, c);
			if ( cur_node == NULL )
				return 0;
		}else if ( (*cur_node)->type->owner->cost >
				c->match->owner->cost ) {
			struct detect *new;

			new = new_node(c);
			if ( new == NULL )
				return 0;

			new->any = *cur_node;

			*cur_node = new;
			cur_node = new->type->insert_node(new, c);
			if ( cur_node == NULL )
				return 0;
		}else if ( is_compatible(*cur_node, c) ) {
			struct detect *n = *cur_node;
			cur_node = c->match->insert_node(n, c);
			if ( cur_node == NULL )
				return 0;
		}else{
			cur_node = &(*cur_node)->any;
			goto do_something;
		}
	}

	free(sig);
	return 1;
}

static void detect_do_free_sig(struct sig_node *sig, unsigned int n)
{
	unsigned int i;

	for(i=0; i < n; i++)
		sig[i].match->free_priv(sig[i].private);

	free(sig);
}

void detect_free_sig(struct sig_node *sig, unsigned int n)
{
	detect_do_free_sig(sig, n);
}

static int is_ignored(struct matcher *m, const char **ignore)
{
	if ( ignore == NULL )
		return 0;
	for(; *ignore; ignore++) {
		if ( !strcmp(m->owner->name, *ignore) )
			return 1;
	}
	return 0;
}

struct sig_node *detect_new_sig(struct rule *r, const char **ignore)
{
	struct criteria *c;
	struct sig_node *x;
	unsigned int i;

	x = calloc(r->num_criteria, sizeof(*x));
	if ( x == NULL )
		return x;

	for(i=0,c=r->criteria; i<r->num_criteria; i++, c=c->next) {
		if ( is_ignored(c->match, ignore) ) {
			r->num_criteria--;
			i--;
			continue;
		}

		x[i].match = c->match;
		x[i].negate = c->negate;

		if ( !c->match->validate(c->match, c->args, &x[i].private,
						c->modifier) ) {
			mesg(M_WARN, "%s: validate failed: '%s'",
				c->match->owner->name, c->args);
			goto fail;
		}
	}

	return realloc(x, r->num_criteria * sizeof(*x));
fail:
	detect_do_free_sig(x, i);
	return 0;
}

void detect(struct detect_root *root, struct packet *p, int l)
{
	if ( root->child == NULL )
		return;

	root->child->match(root->child, p, l);
}
