/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002,2003,2004 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This file is the entry point of the firecat;
* it is responsible for initialising all other components.
*/
#include <firestorm.h>
#include <time.h>
#include <stdarg.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>

#include <mesg.h>
#include <args.h>
#include <alert.h>
#include <loader.h>
#include <signature.h>
#include <decode.h>
#include <capdev.h>
#include <preproc.h>
#include <detect.h>
#include <sensor.h>
#include <exporter.h>
#include <nbio.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

static char *cap[128];
static int num_caps;
static char *output[128];
static int num_outputs;

#ifdef HAVE_SIGACTION
static void signalfn(int signum, siginfo_t *info, void *foo)
#else
#warning "Your architecture doesn't have sigaction, uh oh..."
static void signalfn(int signum)
#endif
{
	const char *action;
	int skipmsg = 0;
	bitmask_t type;

	switch(signum) {
	case SIGKILL:
	case SIGTERM:
	case SIGINT:
		action = "cleaning up";
		type = PIPELINE_INTR_DEADLY;
		break;
	case SIGHUP:
		action = "rotating logs";
		type = PIPELINE_INTR_HUP;
		break;
	case SIGUSR1:
		action = "user defined 1";
		type = PIPELINE_INTR_USR1;
		break;
	case SIGUSR2:
		action = "user defined 2";
		type = PIPELINE_INTR_USR2;
		break;
	case SIGALRM:
		action = "interval timer";
		type = PIPELINE_INTR_ITIMER;
		skipmsg = 1;
		break;
#ifdef SIGVTALRM
	case SIGVTALRM:
		action = "virtual timer";
		type = PIPELINE_INTR_VTIMER;
		skipmsg = 1;
		break;
#endif
	default:
		action = "spurious";
		type = PIPELINE_INTR_SPURIOUS;
		break;
	}

	if ( skipmsg )
		goto do_the_business;

#if 0
#ifdef HAVE_SIGACTION
	mesg(M_INFO,"signal: Got signal %u from pid=%u uid=%u: %s",
		signum, info->si_pid, info->si_uid, action);
#else
	mesg(M_INFO,"signal: Got signal %u: %s", signum, action);
#endif
#endif

do_the_business:
	pipeline_intr(type, 0);
}

#ifdef HAVE_SIGACTION
static sigset_t competing_signals;
static void init_competing_signals(void)
{
	sigemptyset(&competing_signals);
	sigaddset(&competing_signals, SIGINT);
	sigaddset(&competing_signals, SIGTERM);
	sigaddset(&competing_signals, SIGUSR1);
	sigaddset(&competing_signals, SIGUSR2);
	sigaddset(&competing_signals, SIGALRM);
#ifdef SIGVTALRM
	sigaddset(&competing_signals, SIGVTALRM);
#endif
	sigaddset(&competing_signals, SIGHUP);
}
#else
static void init_competing_signals(void)
{
}
#endif

#if HAVE_SIGPROCMASK
static void mask_competing_signals(void *p)
{
	sigprocmask(SIG_BLOCK, &competing_signals, NULL);
}

static void unmask_competing_signals(void *p)
{
	sigprocmask(SIG_UNBLOCK, &competing_signals, NULL);
}
#else
#warning "your OS has no sigprocmask, signal handling will be innacurate"
static void mask_competing_signals(void *p)
{
}
static void unmask_competing_signals(void *p)
{
}
#endif

static struct intr_ops intr_ops = {
	.lock = mask_competing_signals,
	.unlock = unmask_competing_signals,
};

#ifdef HAVE_GETOPT_LONG
static const struct option opts[]={
	{"plugin-dir", 	1, 0, 'p'},
	{"capture", 1, 0, 'c'},
	{"output", 1, 0, 'o'},
	{"merge", 0, 0, 'm'},
	{"verbose", 0, 0, 'v'},
	{"help", 0, 0, 'h'},
	{NULL, 0, 0, 0},
};
#endif
#define MY_OPTS "p:c:o:mvh?"

static void get_plugins(void)
{
#if 0
	loader_load_dir(PLUGIN_DIR "/protocols");
	loader_load_dir(PLUGIN_DIR "/output");
#else
	loader_load_dir("decode_plugins/.libs/");
	loader_load_dir("capdev_plugins/.libs/");
	loader_load_dir("exporter_plugins/.libs/");
#endif
	decode_load();
	eventloop_load();
	capdev_load();
	exporter_load();
	preproc_load();
}

static const char *cmd;
static int verbose_flag;
static void mesg_backend(mesg_code_t code, const char *str, size_t len)
{
	if ( !verbose_flag && code > M_WARN )
		return;

	/* Use stderr cos output may use stdout */
	fprintf(stderr, "%s: %s\n", cmd, str);
}

static void usage(int ecode) __noreturn;
static void usage(int ecode)
{
	extern struct exporter *exporters;
	extern struct capdev *capdevs;
	struct exporter *e;
	struct capdev *c;

	fprintf(stderr, "Usage: %s -c \"CAPDEV [ARGS...]\" [OPTS...]\n", cmd);
	fprintf(stderr, "Generic packet processing utility.\n\n");
	fprintf(stderr, "    --capture, -c     Select a capture source\n");
	fprintf(stderr, "    --output, -o      Select an output type\n");
	fprintf(stderr, "    --merge, -m       Merge input files\n");
	fprintf(stderr, "    --plugin-dir, -p  Load extra plugins from this directory\n");
	fprintf(stderr, "    --verbose, -v     Display debugging messages to stderr\n");
	fprintf(stderr, "    --help, -h        This message\n\n");
	fprintf(stderr, "Loading plugins from: %s\n", PLUGIN_DIR "/protocols");
	fprintf(stderr, "Loading plugins from: %s\n", PLUGIN_DIR "/output");

	get_plugins();

	fprintf(stderr, "\nAvailable capture plugins and their options:\n");
	for(c=capdevs; c; c=c->next) {
		struct arg *a;

		fprintf(stderr, "  %s [%s] - %s\n",
			c->name,
			(c->flags & CAPDEV_NBIO) ? "live" : "file",
			c->desc);
		if ( c->args == NULL )
			continue;

		for(a = c->args; a->name; a++)
			fprintf(stderr, "   o %s: %s (%s%s%s)\n",
				a->name,
				a->help ? a->help : "[no help available]",
				args_typename(a->type),
				a->def ? " default=" : "",
				a->def ? a->def : "");
	}

	fprintf(stderr, "\nAvailable output plugins and their options:\n");
	for(e=exporters; e; e=e->next) {
		struct arg *a;

		fprintf(stderr, "  %s - %s\n", e->name, e->desc);
		if ( e->args == NULL )
			continue;

		for(a = e->args; a->name; a++)
			fprintf(stderr, "   o %s: %s (%s%s%s)\n",
				a->name,
				a->help ? a->help : "[no help available]",
				args_typename(a->type),
				a->def ? " default=" : "",
				a->def ? a->def : "");
	}

	fprintf(stderr, "\nReport bugs to <gianni@scaramanga.co.uk>\n");
	firestorm_exit(ecode, NULL);
}

static const char *strip_dir(const char *fn)
{
	const char *ret;

	for(ret = fn; *fn != '\0'; fn++) {
		if ( *fn == '/' && (*(fn + 1) != '\0') )
			ret = fn + 1;
	}

	return ret;
}

int main(int argc, char **argv)
{
	struct pipeline p;
	int c;
	bitmask_t p_flags = PIPELINE_ALLOW_LIVE|
				PIPELINE_ALLOW_FILE|
				PIPELINE_SINK_ERRORS_FATAL;
#ifdef HAVE_SIGACTION
	struct sigaction sig;
#endif

#ifdef HAVE_TZSET
	/* Set timezone from locale */
	tzset();
#endif
	cmd = strip_dir(argv[0]);

	/* Setup printing to screen */
	mesg_set_backend(mesg_backend);

	firestorm_init();

	/* Print copyright information */
	mesg(M_INFO,"Firestorm Sniffer v" PACKAGE_VERSION);
	mesg(M_INFO,"Copyright (c) 2002-2004 Gianni Tedesco");
	mesg(M_INFO,"This program is free software; released under "
		"the GNU GPL v2 (see: COPYING)");

#if HAVE_GETOPT_LONG
	while ( (c=getopt_long(argc, argv, MY_OPTS, opts, NULL)) != -1 )
#else
	while ( (c=getopt(argc, argv, MY_OPTS)) != -1 )
#endif
	{
		switch ( c ) {
		case 'c':
			cap[num_caps++] = optarg;
			break;
		case 'o':
			output[num_outputs++] = optarg;
			break;
		case 'm':
			p_flags |= PIPELINE_MERGE;
			break;
		case 'p':
			loader_load_dir(optarg);
			break;
		case 'v':
			verbose_flag = 1;
			break;
		case 'h':
			usage(EXIT_SUCCESS);
		default:
			usage(EXIT_FAILURE);
		}
	}

	get_plugins();

	if ( !pipeline_init(&p, p_flags) )
		firestorm_exit(EXIT_FAILURE, "pipeline_init failed");

	init_competing_signals();
	pipeline_intr_ops(&intr_ops, NULL);

#if 0
	/* Use some basic preprocessors */
	if ( !(preproc_setup("ipfrag",
			"mem_hi=4096k mem_lo=3072k minttl=0 timeout=30") &&
		preproc_setup("tcpstream",
			"num_streams=32k num_flows=32k reassemble=yes") && 
		preproc_setup("http_normalize",
			"emulate=IIS")) ) {
		mesg(M_DEBUG, "preprocessors failed to load");
	}
#endif

	if ( num_caps == 0 ) {
		firestorm_exit(EXIT_FAILURE, "no capture source specified");
	}

	for(c=0; c < num_caps; c++) {
		char *tok[2] = {NULL,NULL};
		easy_explode(cap[c], '\0', tok, 2);
		if ( !pipeline_source_add(&p, tok[0], tok[1]) ) {
			firestorm_exit(EXIT_FAILURE,
				"capture: %s[%s] failed to load",
				tok[0], tok[1]);
		}
	}

	for(c=0; c < num_outputs; c++) {
		char *tok[2] = {NULL,NULL};
		easy_explode(output[c], '\0', tok, 2);
		if ( !pipeline_sink_add(&p, SINK_REAL, tok[0], tok[1]) ) {
			firestorm_exit(EXIT_FAILURE,
				"output: %s[%s] failed to load",
				tok[0], tok[1]);
		}
	}

	/* Catch INT and TERM for clean exits */
	/* Catch HUP to rotate logs */
	/* Catch USR1 and USR2 to use later */
	/* Catch SIGALRM for itimer for logrotates */
	/* Ignore SIGPIPE (for stormwall fifo) */
#ifdef HAVE_SIGACTION
	sig.sa_sigaction=signalfn;
	sig.sa_flags=SA_SIGINFO;
	sig.sa_mask = competing_signals;
	sigaddset(&sig.sa_mask, SIGINT);
	sigaddset(&sig.sa_mask, SIGTERM);
	sigaddset(&sig.sa_mask, SIGUSR1);
	sigaddset(&sig.sa_mask, SIGUSR2);
	sigaddset(&sig.sa_mask, SIGALRM);
	sigaddset(&sig.sa_mask, SIGHUP);
	sigaction(SIGINT, &sig, NULL);
	sigaction(SIGTERM, &sig, NULL);
	sigaction(SIGUSR1, &sig, NULL);
	sigaction(SIGUSR2, &sig, NULL);
	sigaction(SIGALRM, &sig, NULL);
	sigaction(SIGHUP, &sig, NULL);

	sig.sa_handler=SIG_IGN;
	sig.sa_flags=SA_SIGINFO;
	sigemptyset(&sig.sa_mask);
	sigaction(SIGPIPE, &sig, NULL);
#else
	signal(SIGINT, signalfn);
	signal(SIGTERM, signalfn);
	signal(SIGUSR1, signalfn);
	signal(SIGUSR2, signalfn);
	signal(SIGALRM, signalfn);
	signal(SIGHUP, signalfn);

	signal(SIGPIPE, SIG_IGN);
#endif

	pipeline_go(&p);

	pipeline_free(&p);

	firestorm_exit(EXIT_SUCCESS, "Firestorm exiting normally");
}
