/*
* This file is part of Firestorm NIDS
* Copyright (c) 2003,2004 Gianni Tedesco
* Released under the terms of the GNU GPL version 2
*
* This code takes a bunch of i_ordered databses and squeezes them all
* out in to one big database file. On-disk formats are described in
* idx.h.
*/

#include <firestorm.h>
#include <dotscara/fobuf.h>
#include <alert.h>
#include <idx.h>
#include <i_idx.h>

static struct i_type *idx_types[IDX_TYPES]={
	[ IDX_SKUNK ] &idx_skunk,
	[ IDX_SKUNK_NOCASE ] &idx_skunk_nocase,
	[ IDX_ORDERED_UINT8 ] &idx_ordered_uint32,
	[ IDX_ORDERED_UINT16 ] &idx_ordered_uint32,
	[ IDX_ORDERED_UINT32 ] &idx_ordered_uint32,
};

struct i_idx *i_idx_new(void)
{
	struct i_idx *ret;

	ret = malloc(sizeof(*ret));
	if ( !ret )
		return NULL;

	ret->num_fields = 0;
	ret->fields = NULL;

	return ret;
}

struct i_fieldref *i_idx_add_field(struct i_idx *idx, int type, const char *name)
{
	struct i_fieldref *f;
	struct i_type *t;

	if ( type < 0 || type > IDX_TYPES )
		return NULL;

	if ( strlen(name)+1 > IDX_NAMELEN )
		return NULL;

	t = idx_types[type];
	if ( t == NULL )
		return NULL;

	f = t->new();
	if ( f == NULL )
		return NULL;

	f->type = t;
	f->name = name;
	f->next = idx->fields;
	idx->fields = f;
	idx->num_fields++;

	return f;
}

int i_idx_write_fd(struct i_idx *idx, int fd)
{
	struct fobuf o;
	struct idx_hdr hdr;
	struct i_fieldref *f;
	idx_ofs_t ofs;
	int ret = 1;

	/* Use a 32KB output buffer */
	if ( !fobuf_attach(&o, fd, 0, 32768) )
		return 0;

	/* write header */
	hdr.magic = IDX_MAGIC;
	hdr.num = idx->num_fields;
	if ( !fobuf_write(&o, &hdr, sizeof(hdr)) ) {
		fobuf_close(&o);
		return 0;
	}

	ofs = sizeof(hdr) + hdr.num * sizeof(struct idx_index);

	/* Write the field descriptors */
	for(f=idx->fields; f; f=f->next) {
		struct idx_index db;

		memset(db.name, '\0', sizeof(db.name));
		snprintf(db.name, sizeof(db.name), "%s", f->name);

		db.type = f->type->id;
		db.reserved1 = 0;
		db.reserved2 = 0;
		db.ofs = ofs;

		if ( !f->type->prepare(f, &db.len,
					&db.num_keys, &db.num_vals) ) {
			fobuf_close(&o);
			return 0;
		}

		if ( !fobuf_write(&o, &db, sizeof(db)) ) {
			fobuf_close(&o);
			return 0;
		}
		ofs += db.len;
	}
	
	/* Write the actual databases */
	for(f=idx->fields; f; f=f->next) {
		if ( !f->type->write(f, &o) ) {
			ret = 0;
			break;
		}

		f->type->free(f);
	}

	/* Free them all, even in error path */
	for(; f; f=f->next)
		f->type->free(f);

	/* Sync it to disk */
	if ( !fobuf_close(&o) )
		ret = 0;

	return ret;
}

void i_idx_free(struct i_idx *idx)
{
	free(idx);
}
