/*
* This file is part of Firestorm NIDS
* Copyright (c) 2003,2004 Gianni Tedesco
* Released under the terms of the GNU GPL version 2
*
* avl tree maintenance functions taken from avl-0.4, original copyright
* notice is reproduced below:
*
* (C) 2000 Daniel Nagy, Budapest University of Technology and Economics
* Released under GNU General Public License (GPL) version 2
*
* TODO:
*  o Implement the smaller sizes
*/
#include <firestorm.h>
#include <dotscara/fobuf.h>
#include <dotscara/mpool.h>
#include <idx.h>
#include <i_idx.h>

struct i_ordered {
	struct i_fieldref f;
	struct mpool avl_nodes;
	struct avl *cache;
	struct avl *root;
	size_t num_keys;
	size_t num_vals;
};

struct avl {
	struct avl *left, *right;
	idx_key_t key;
	int balance;
	struct mpool vals;
	idx_ofs_t num_vals;
};

static struct avl *cur;

/* avl_swl:
 *
 * AVL tree maintenance function
*/
static inline void avl_swl(struct avl **r)
{
	struct avl *a=*r;
	struct avl *b=a->right;
	*r=b;
	a->right=b->left;
	b->left=a;
}

/* avl_swr:
 *
 * AVL tree maintenance function
*/
static inline void avl_swr(struct avl **r)
{
	struct avl *a=*r;
	struct avl *b=a->left;
	*r=b;
	a->left=b->right;
	b->right=a;
}

/* avl_balance:
 *
 * AVL tree maintenance function
*/
static inline void avl_balance(struct avl *r)
{
	switch ( r->balance ) {
	case -1:
		r->left->balance=0;
		r->right->balance=1;
		break;
	case 1:
		r->left->balance=-1;
		r->right->balance=0;
		break;
	case 0:
		r->left->balance=0;
		r->right->balance=0;
	}
	r->balance=0;
}

static inline int hdr_count(struct mpool *m)
{
	int ret=0;
	struct mpool_hdr *h;

	for(h=m->slabs; h; h=h->next)
		ret++;

	return ret;
}

static void reverse_mpool(struct mpool *m)
{
	int cnt=hdr_count(m);
	struct mpool_hdr *a[cnt];
	struct mpool_hdr *h;
	int i=0;

	for(h=m->slabs; h; h=h->next)
		a[i++] = h;

	m->slabs=NULL;

	for(i=0; i<cnt; i++) {
		a[i]->next = m->slabs;
		m->slabs = a[i];
	}
}

/* avl_insert_key:
 * @r: double-indirect pointer to the root #avl node
 * @n: the node to insert
 *
 * Insert an AVL node in to the tree, sets the global variable cur to be
 * a pointer to the unique node to use. If there is already a node for
 * that key in the tree @n will not be used.
 *
 * Return value: for recursion only, to be ignored.
 */
static int avl_insert_key(struct avl **r, struct avl *n)
{
	struct avl *t=*r;

	if ( n->key < t->key ) {
		if ( t->left ) {
			if ( avl_insert_key(&t->left, n) ) {
				switch ( t->left->balance-- ) {
				case 1: return 0;
				case 0: return 1;
				}
				if ( t->left->balance<0 ) {
					/* simple rotate */
					avl_swr(r);
					(*r)->balance=0;
					(*r)->right->balance=0;
				}else{
					/* double rotate */
					avl_swl(&t->left);
					avl_swr(r);
					avl_balance(*r);
				}
			}
			return 0;
		}else{
			t->left=n;
			if ( t->balance-- ) return 0;
			return 1;
		}
	}

	if ( n->key > t->key ) {
		if ( t->right ) {
			if ( avl_insert_key(&t->right, n) ) {
				switch ( t->right->balance-- ) {
				case 1: return 0;
				case 0: return 1;
				}
				if ( t->right->balance<0 ) {
					/* simple rotate */
					avl_swl(r);
					(*r)->balance=0;
					(*r)->left->balance=0;
				}else{
					/* double rotate */
					avl_swr(&t->right);
					avl_swl(r);
					avl_balance(*r);
				}
			}
			return 0;
		}else{
			t->right=n;
			if ( t->balance-- ) return 0;
			return 1;
		}
	}

	if ( n->key==t->key )
		cur=t;

	return 0;
}

/* ordered_insert:
 * @db: a valid #i_ordered database
 * @key: key to use
 * @val: offset in file being indexed
 *
 * Inserts an item in to the index.
 *
 * Return value: zero on error, non-zero on success.
 */
static u_int32_t vals;
static int ordered_insert(struct i_fieldref *f, idx_key_t key, idx_val_t val)
{
	struct i_ordered *db = (struct i_ordered *)f;
	struct avl *n;
	int new_key=0;
	idx_val_t *v;

	if ( db->cache ) {
		n=db->cache;
	}else{
		n=db->cache=mpool_alloc(&db->avl_nodes);
		if ( n == NULL )
			return 0;
	}

	n->left=NULL;
	n->right=NULL;
	n->key=key;
	n->balance=0;
	n->num_vals=0;
	cur=n;

	if ( db->root ) {
		avl_insert_key(&db->root, n);
		if ( cur==n )
			new_key=1;
	}else{
		db->root=n;
		new_key=1;
	}

	if ( new_key ) {
		if ( !mpool_new(&cur->vals, sizeof(idx_val_t), 128) )
			return 0;

		db->num_keys++;
		cur->num_vals=0;
		db->cache=NULL;
	}

	/* Add the value */
	if ( !(v=mpool_alloc(&cur->vals)) ) {
		mesg(M_CRIT, "%u vals", vals);
		return 0;
	}

	vals++;

	*v=val;
	db->num_vals++;
	cur->num_vals++;

	return 1;
}

static void i_ordered_free_vals(struct avl *n)
{
	if ( n->left )
		i_ordered_free_vals(n->left);

	mpool_free(&n->vals);

	if ( n->right )
		i_ordered_free_vals(n->right);
}

static idx_ofs_t cur_ofs;

static int i_ordered_write_key(struct avl *n, struct fobuf *o)
{
	if ( n->left ) {
		if ( !i_ordered_write_key(n->left, o) )
			return 0;
	}

	do {
		struct idx_onode dnode;

		dnode.key = n->key;
		dnode.ofs = cur_ofs;
		dnode.num = n->num_vals;

		cur_ofs += sizeof(idx_val_t) * n->num_vals;

		if ( !fobuf_write(o, &dnode, sizeof(dnode)) )
			return 0;
	}while(0);

	if ( n->right ) {
		if ( !i_ordered_write_key(n->right, o) )
			return 0;
	}

	return 1;
}

static int i_ordered_write_val(struct avl *n, struct fobuf *o)
{
	if ( n->left ) {
		if ( !i_ordered_write_val(n->left, o) )
			return 0;
	}

	do {
		struct mpool_hdr *h;
		size_t tlen=0;

		reverse_mpool(&n->vals);

		for(h=n->vals.slabs; h; h=h->next) {
			size_t len;

			/* FIXME: once we add alignment and/or coloring
			 * this will fail badly */
			len = (char *)h->next_obj - (char *)h;
			if ( len > n->vals.slab_size )
				len -= n->vals.obj_size;

			len -= sizeof(*h);

			tlen += len;

			if ( !fobuf_write(o, h->data, len) )
				return 0;
		}

		/* to catch errors iterating the mpool */
		if ( tlen != n->num_vals*sizeof(idx_val_t) )
			return 0;
	}while(0);

	if ( n->right ) {
		if ( !i_ordered_write_val(n->right, o) )
			return 0;
	}

	return 1;
}

/* ordered_write:
 * @db: a valid #i_ordered database
 * @o: an output file buffer
 *
 * Writes the database out to the file
 *
 * Return value: zero on failure, non-zero on success
 */
static int ordered_write(struct i_fieldref *f, struct fobuf *o)
{
	struct i_ordered *db = (struct i_ordered *)f;

	cur_ofs = db->num_keys * sizeof(struct idx_onode);

	if ( !db->root )
		return 1;

	if ( !i_ordered_write_key(db->root, o) )
		return 0;

	if ( !i_ordered_write_val(db->root, o) )
		return 0;

	return 1;
}

/* ordered_new:
 *
 * Creates a new DB for inserting in to.
 *
 * Return value: NULL on error
 */
static struct i_fieldref *ordered_new(void)
{
	struct i_ordered *db;

	db = calloc(1, sizeof(*db));
	if ( db == NULL )
		return NULL;

	/* Create memory allocator for AVL tree nodes */
	if ( !mpool_new(&db->avl_nodes, sizeof(struct avl), 0) ) {
		free(db);
		return NULL;
	}

	db->cache=NULL;
	db->root=NULL;
	db->num_keys=0;
	db->num_vals=0;

	return (struct i_fieldref *)db;
}

/* ordered_free:
 * @db: a valid #i_ordered database
 *
 * Deletes all data currently associated with the database. The
 */
static void ordered_free(struct i_fieldref *f)
{
	struct i_ordered *db = (struct i_ordered *)f;

	if ( db->root )
		i_ordered_free_vals(db->root);

	mpool_free(&db->avl_nodes);
	db->num_keys=0;
	db->num_vals=0;
	db->cache=NULL;
}

/* ordered_prepare:
 * @db: a valid #i_ordered database
 * @size: pointer to size of total DB
 * @keys: pointer to number of keys
 * @vals: pointer to number of vals
 *
 * This method is called once after all inserts have been done, but
 * before writing. The size, keys and vals params must all be filled
 * in in a valid way.
 *
 * Return value: zero on success, less than zero for error.
 */
static int ordered_prepare(struct i_fieldref *f, idx_ofs_t *size,
			idx_ofs_t *keys, idx_ofs_t *vals)
{
	struct i_ordered *db = (struct i_ordered *)f;
	*size = (db->num_keys * sizeof(struct idx_onode)) +
		(db->num_vals * sizeof(idx_val_t));
	*keys = db->num_keys;
	*vals = db->num_vals;
	return 1;
}

/* Ordered unsigned 32bit int type */
struct i_type idx_ordered_uint32 __private = {
	.id = IDX_ORDERED_UINT32,
	.new = ordered_new,
	.insert = ordered_insert,
	.insert_str = NULL,
	.prepare = ordered_prepare,
	.write = ordered_write,
	.free = ordered_free,
};
