/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002,2003,2004 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This is the firestorm plugin loader. We can load individually
* named plugins or scan directories for them. We don't recurse
* directories. We also perform all the plugin initialisation
* and unloading from here. Plugin unloading is the last thing
* to happen before exiting
*
* Functions:
*  o loader_cleanup() - unload all plugins
*  o loader_load() - load a plugin from path
*  o loader_load_dir() - load all plugins in a directory
*  o loader_find_syms() - find all plugins exporting a given symbol
*/
#include <firestorm.h>
#include <loader.h>

#ifdef HAVE_DIRENT_H
#include <dirent.h>
#endif
#ifdef HAVE_NDIR_H
#include <ndir.h>
#endif
#ifdef HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#ifdef HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif

#ifdef HAVE_DLFCN_H
#include <dlfcn.h>
#else
#warning "Can't find dlfcn.h, this should be part of your libc"
#endif

#ifdef RTLD_NOW
#define LINK_MODE RTLD_NOW
#else
#define LINK_MODE RTLD_LAZY
#endif

static struct plugin *plugin_list=NULL;

static void loader_cleanup(struct packet *pkt, int l, void *priv)
{
	struct plugin *foo, *bar;

	mesg(M_INFO,"loader: unloading all plugins");

	for(foo=plugin_list; foo;) {
		bar=foo;
		foo=foo->next;
		if ( bar->d->fini )
			bar->d->fini();
		dlclose(bar->lib);
		free(bar);
	}
}

void __private loader_init(void)
{
	notify_exit(loader_cleanup, NULL);
}

/* Load a plugin given a filename */
int loader_load(const char *path)
{
	struct plugin *p;

	/* Allocate private structure */
	if ( !(p=calloc(1, sizeof(struct plugin))) ) {
		firestorm_die_error("calloc");
	}

	/* dlopen() the library */
	if ( !(p->lib=dlopen(path, LINK_MODE)) ) {
		mesg(M_ERR,"loader: %s", dlerror());
		goto llerr;
	}

	/* Get the descriptor block */
	p->d = dlsym(p->lib, "firestorm_plugin_desc");
	if ( p->d == NULL ) {
		mesg(M_ERR, "loader: %s: plugin descriptor not found", path);
		goto llerr;
	}

	if ( p->d->size != sizeof(*p->d) ) {
		mesg(M_ERR, "loader: %s: invalid plugin descriptor", path);
		goto llerr;
	}

	/* Run the init function */
	if ( p->d->init )
		p->d->init();

#if 0
	mesg(M_INFO, "plugin: %s[%i.%i]: %s",
			p->d->name,
			p->d->ver_major,
			p->d->ver_minor,
			p->d->desc);
#endif

	p->next = plugin_list;
	plugin_list = p;

	return 1;

llerr:
	if ( p->lib )
		dlclose(p->lib);
	free(p);
	return 0;
}

/* Iterate all files in a directory and try
 * to load them as plugins, oh what folly... */
void loader_load_dir(const char *dir)
{
	DIR *d;
	struct dirent *f;
	char fn[1024]; /* XXX: Should be >MAX_PATH on all platforms */

	if ( !(d=opendir(dir)) ) {
		mesg(M_ERR,"loader: %s: opendir(): %s",
			dir, sys_err());
		return;
	}

	while( (f=readdir(d)) ) {
		int len;

		/* Don't load hidden files or . or .. */
		if ( f->d_name[0] == '.' ) continue;

		/* only load files which end with .so */
		len=strlen(f->d_name);
		if ( len<3 || strcmp(f->d_name+len-3, ".so") )
			continue;

		if ( snprintf(fn, sizeof(fn), "%s/%s", dir, f->d_name)>0 ) {
			loader_load(fn);
		}else{
			mesg(M_ERR,"loader: %s: path overflow.", dir);
		}
	}

	closedir(d);
}

void loader_find_syms(const char *sym, loader_cb_t cb)
{
	struct plugin *p;
	void *s;

	for(p=plugin_list; p; p=p->next) {
		s = dlsym(p->lib, sym);
		if ( s == NULL )
			continue;

		(*cb)(p, s);
	}
}
