/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002,2003,2004 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* Some template parsing functions for use by matcher plugins.
*/
#include <firestorm.h>

#include <errno.h>

#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <detect.h>
#include <idx.h>
#include <field.h>
#include <matcher.h>

static inline int strtoulong(const char *s, unsigned long *v)
	__nonull(1,2);
static inline int strtoulong(const char *s, unsigned long *v)
{
	unsigned long int ret;
	char *end=NULL;

	ret=strtoul(s, &end, 0);

	if ( end==s )
		return -1;

	if ( ret == ULONG_MAX && errno==ERANGE )
		return -1;

	*v = ret;

	if ( *end == '\0' )
		return 0;

	return end - s;
}

static int p_range(char *str, unsigned long *min,
			unsigned long *max, unsigned long mask)
{
	unsigned long num;
	char *lo=NULL, *hi=NULL;
	int state=0;
	char *tmp;
	int eq = 0;

	/* Trim leading whitespace */
	while ( isspace(*str) )
		str++;

	/* Handle < > <= >= */
	if ( *str == '<' ) {
		if ( *(++str) == '=' ) {
			eq = 1;
			str++;
		}
		if ( strtoulong(str, &num) )
			return 0;
		if ( !eq )
			num--;
		*min = 0;
		*max = num;
		return 1;
	}else if ( *str == '>' ) {
		if ( *(++str) == '=' ) {
			eq = 1;
			str++;
		}
		if ( strtoulong(str, &num) )
			return 0;
		if ( !eq )
			num++;
		*min = num;
		*max = mask-1;
		return 1;
	}

	/* Parse the string out */
	for(tmp=str; *tmp; tmp++) {
		if ( state == 0 ) {
			if ( *tmp == ':' ) {
				state = 2;
			}else if ( !isspace(*tmp) ) {
				lo = tmp;
				state = 1;
			}
		}else if ( state == 1 ) {
			if ( isspace(*tmp) ) {
				*tmp = 0;
			}else if ( *tmp == ':' ) {
				state = 2;
				*tmp = 0;
			}
		}else if ( state == 2 ) {
			if ( !isspace(*tmp) ) {
				hi = tmp;
				state = 3;
			}
		}else if ( state == 3 ) {
			if ( isspace(*tmp) ) {
				*tmp = 0;
				break;
			}
		}
	}

	/* Low part is easy */
	if ( lo && *lo ) {
		if ( strtoulong(lo, &num) )
			return 0;
		if ( num & ~mask )
			return 0;
		*min = num & mask;
	} else
		*min = 0;

	if ( hi ) {
		if ( strtoulong(hi, &num) )
			return 0;
		if ( num & ~mask )
			return 0;
		*max = num & mask;
	}else{
		/* If state is 1 that means there was a colon
		 * eg: 1024: otherwise it was just a single
		 * number, eg: '23' in which case max
		 * is just equal to min
		 */
		if ( state > 1 )
			*max = mask;
		else
			*max = *min;
	}

	return 1;
}

int template_optimize(struct detect *d)
{
	return 1;
}

static void template_match(struct detect *d, struct packet *p, int l)
{
	struct template_detect *x = (struct template_detect *)d;
	struct template_node *n;
	u_int32_t val;

	if ( !x->cbfn(p, l, &val) )
		goto skip;

	for(n=x->nodes; n; n=n->next) {
		if ( n->negate ^ ((val >= n->u.val.min)
				&& (val <= n->u.val.max)) ) {
			alert_tag(p, n->alert, l);
			if ( n->result )
				n->result->match(n->result, p, l);
		}
	}

skip:
	if ( likely(x->d.any) )
		x->d.any->match(x->d.any, p, l);
}

static void template_match_str(struct detect *d, struct packet *p, int l)
{
	struct template_detect2 *x = (struct template_detect2 *)d;
	struct template_node *n;
	struct fvec vec;

	if ( !x->cbfn(p, l, &vec) )
		goto skip;

	for(n=x->nodes; n; n=n->next) {
		if ( !vcmp(&vec, &n->u.str) ) {
			alert_tag(p, n->alert, l);
			if ( n->result )
				n->result->match(n->result, p, l);
		}
	}

skip:
	if ( likely(x->d.any) )
		x->d.any->match(x->d.any, p, l);
}

static void template_match_str_nocase(struct detect *d, struct packet *p, int l)
{
	struct template_detect2 *x = (struct template_detect2 *)d;
	struct template_node *n;
	struct fvec vec;

	if ( !x->cbfn(p, l, &vec) )
		goto skip;

	for(n=x->nodes; n; n=n->next) {
		if ( !vcasecmp(&vec, &n->u.str) ) {
			alert_tag(p, n->alert, l);
			if ( n->result )
				n->result->match(n->result, p, l);
		}
	}

skip:
	if ( likely(x->d.any) )
		x->d.any->match(x->d.any, p, l);
}

struct detect *template_new_node_str(proc_template_getstr cbfn)
{
	struct template_detect2 *x;

	if ( cbfn == NULL )
		return NULL;

	x = calloc(1, sizeof(*x));
	if ( x == NULL )
		return NULL;

	x->d.match = template_match_str;
	x->cbfn = cbfn;

	return (struct detect *)x;
}

struct detect *template_new_node_str_nocase(proc_template_getstr cbfn)
{
	struct template_detect2 *x;

	if ( cbfn == NULL )
		return NULL;

	x = calloc(1, sizeof(*x));
	if ( x == NULL )
		return NULL;

	x->d.match = template_match_str_nocase;
	x->cbfn = cbfn;

	return (struct detect *)x;
}

struct detect *template_new_node(proc_template_getval cbfn)
{
	struct template_detect *x;

	if ( cbfn == NULL )
		return NULL;

	x = calloc(1, sizeof(*x));
	if ( x == NULL )
		return NULL;

	x->d.match = template_match;
	x->cbfn = cbfn;

	return (struct detect *)x;
}

struct detect **template_insert_default(struct detect *d)
{
	struct template_detect *x = (struct template_detect *)d;
	return &x->d.any;
}

struct detect **template_insert(struct detect *d, struct sig_node *n)
{
	struct template_detect *x = (struct template_detect *)d;
	struct template_node *p = (struct template_node *)n->private;
	struct template_node *tmp;

	p->alert = n->alert;
	p->negate = n->negate;

	for(tmp=x->nodes; tmp; tmp=tmp->next) {
		if ( p->u.val.min != tmp->u.val.min )
			continue;
		if ( p->u.val.max != tmp->u.val.max )
			continue;
		if ( p->negate != tmp->negate )
			continue;

		tmp->alert = alert_best(tmp->alert, p->alert);
		n->match->free_priv(n->private);
		return &tmp->result;
	}

	p->next = x->nodes;
	x->nodes = p;

	return &p->result;
}

int template_validate(struct matcher *m, char *args,
			void **priv, struct modifier *opts)
{
	unsigned long min, max;
	struct fvec vec = {.ptr = NULL};
	struct template_node *n;
	idx_key_t tmp = 0;
	u_int32_t mask;

	if ( m->owner->parse.val ) {
		switch ( m->owner->type ) {
		case IDX_ORDERED_UINT8:
		case IDX_ORDERED_INT8:
		case IDX_ORDERED_UINT16:
		case IDX_ORDERED_INT16:
		case IDX_ORDERED_UINT32:
		case IDX_ORDERED_INT32:
			if ( !m->owner->parse.val(args, &tmp) )
				break;
			min = max = (unsigned long)tmp;
			goto got_val;
		case IDX_SKUNK:
		case IDX_SKUNK_NOCASE:
			if ( !m->owner->parse.str(args, &vec) )
				break;
			goto got_str;
		default:
			return 0;
		}
	}

	switch( m->owner->type ) {
	case IDX_ORDERED_UINT8:
	case IDX_ORDERED_INT8:
		mask = 0xffUL;
		break;
	case IDX_ORDERED_UINT16:
	case IDX_ORDERED_INT16:
		mask = 0xffffUL;
		break;
	case IDX_ORDERED_UINT32:
	case IDX_ORDERED_INT32:
		mask = 0xffffffffUL;
		break;
	case IDX_SKUNK:
	case IDX_SKUNK_NOCASE:
		vec.len = strlen(args);
		if ( vec.len ) {
			vec.ptr = malloc(vec.len);
			if ( vec.ptr == NULL )
				firestorm_die_error("malloc");
			memcpy(vec.ptr, args, vec.len);
		}
		goto got_str;
	default:
		return 0;
	}

	if ( !p_range(args, &min, &max, mask) )
		return 0;

got_val:
	n = calloc(1, sizeof(*n));
	if ( n == NULL )
		firestorm_die_error("calloc");

	n->u.val.min = min;
	n->u.val.max = max;
	*priv = n;
	return 1;

got_str:
	n = calloc(1, sizeof(*n));
	if ( n == NULL )
		firestorm_die_error("calloc");

	n->u.str.ptr = vec.ptr;
	n->u.str.len = vec.len;

	*priv = n;
	return 1;
}
